#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import time
from datetime import datetime, timedelta

from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class BridgeContext:
    def __init__(self, **kwargs):
        """
        Bridge context

        :param kwargs: keyword arguments
        """
        self.ctx = kwargs.get("ctx", CtxItem())
        self.history = kwargs.get("history", [])
        self.mode = kwargs.get("mode", None)
        self.parent_mode = kwargs.get("parent_mode", None)  # real mode (global)
        self.model = kwargs.get("model", None)  # model instance, not model name
        self.temperature = kwargs.get("temperature", 1.0)
        self.prompt = kwargs.get("prompt", "")
        self.system_prompt = kwargs.get("system_prompt", "")
        self.system_prompt_raw = kwargs.get("system_prompt_raw", "")  # without plugins addons
        self.stream = kwargs.get("stream", False)
        self.assistant_id = kwargs.get("assistant_id", "")
        self.thread_id = kwargs.get("thread_id", "")
        self.external_functions = kwargs.get("external_functions", [])
        self.tools_outputs = kwargs.get("tools_outputs", [])
        self.max_tokens = kwargs.get("max_tokens", 150)
        self.idx = kwargs.get("idx", "base")
        self.idx_raw = kwargs.get("idx_raw", False)
        self.attachments = kwargs.get("attachments", [])
        self.file_ids = kwargs.get("file_ids", [])

        # check types
        if self.ctx is not None and not isinstance(self.ctx, CtxItem):
            raise ValueError("Invalid context instance")
        if self.model is not None and not isinstance(self.model, ModelItem):
            raise ValueError("Invalid model instance")

    def to_dict(self) -> dict:
        """
        Return as dictionary

        :return: dictionary
        """
        return {
            "ctx": self.ctx,  # "ctx": self.ctx.dump() ??
            "history": self.history,
            "mode": self.mode,
            "parent_mode": self.parent_mode,
            "model": self.model,
            "temperature": self.temperature,
            "prompt": self.prompt,
            "system_prompt": self.system_prompt,
            "system_prompt_raw": self.system_prompt_raw,
            "stream": self.stream,
            "assistant_id": self.assistant_id,
            "thread_id": self.thread_id,
            "external_functions": self.external_functions,
            "tools_outputs": self.tools_outputs,
            "max_tokens": self.max_tokens,
            "idx": self.idx,
            "idx_raw": self.idx_raw,
            "attachments": self.attachments,
            "file_ids": self.file_ids,
        }


class Bridge:
    def __init__(self, window=None):
        """
        Provider bridge

        :param window: Window instance
        """
        self.window = window
        self.last_call = None  # last API call time, for throttling

    def call(self, context: BridgeContext, extra: dict = None) -> bool:
        """
        Make call to provider

        :param context: Bridge context
        :param extra: extra data  # TODO: object also
        """
        allowed_model_change = ["vision"]

        self.window.stateChanged.emit(self.window.STATE_BUSY)  # set busy

        # debug
        self.window.core.debug.info("Bridge call...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))

        # get data
        ctx = context.ctx
        prompt = context.prompt
        mode = context.mode
        model = context.model  # model instance

        # get agent internal sub-mode
        if mode == "agent":
            mode = "chat"  # inline switch to sub-mode, because agent is a virtual mode only
            sub_mode = self.window.core.agents.get_mode()
            if sub_mode is not None and sub_mode != "_":
                mode = sub_mode
            if mode == "llama_index":
                context.idx_raw = False
                idx = self.window.core.agents.get_mode()
                if idx is not None and idx != "_":
                    context.idx = idx

        # inline: internal mode switch if needed
        context.parent_mode = mode  # store REAL mode
        mode = self.window.controller.mode.switch_inline(mode, ctx, prompt)
        context.mode = mode

        # inline: model switch
        if mode in allowed_model_change:
            model = self.window.controller.model.switch_inline(mode, model)
            context.model = model

        # debug
        self.window.core.debug.info("Bridge call (after inline)...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))

        self.apply_rate_limit()  # apply RPM limit

        # Langchain
        if mode == "langchain":
            return self.window.core.chain.call(
                context=context,
                extra=extra,
            )

        # Llama-index
        elif mode == "llama_index":
            return self.window.core.idx.chat.call(
                context=context,
                extra=extra,
            )

        # OpenAI API: chat, completion, vision, image, assistants, etc.
        else:
            return self.window.core.gpt.call(
                context=context,
                extra=extra,
            )

    def quick_call(self, context: BridgeContext, extra: dict = None) -> str:
        """
        Make quick call to provider and get response content

        :param context: Bridge context
        :param extra: extra data
        :return: response content
        """
        self.window.core.debug.info("Bridge quick call...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))
        return self.window.core.gpt.quick_call(
            context=context,
            extra=extra,
        )

    def apply_rate_limit(self):
        """Apply API calls RPM limit"""
        max_per_minute = 60
        if self.window.core.config.has("max_requests_limit"):
            max_per_minute = int(self.window.core.config.get("max_requests_limit")) # per minute
        if max_per_minute <= 0:
            return
        interval = timedelta(minutes=1) / max_per_minute
        now = datetime.now()
        if self.last_call is not None:
            time_since_last_call = now - self.last_call
            if time_since_last_call < interval:
                sleep_time = (interval - time_since_last_call).total_seconds()
                self.window.core.debug.debug("RPM limit: sleep for {} seconds".format(sleep_time))
                time.sleep(sleep_time)
        self.last_call = now



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This store is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.26 23:00:00                  #
# ================================================== #

import uuid

from packaging.version import Version

from pygpt_net.item.assistant import AssistantStoreItem
from pygpt_net.provider.core.assistant_store.base import BaseProvider

from .patch import Patch
from .storage import Storage


class DbSqliteProvider(BaseProvider):
    def __init__(self, window=None):
        super(DbSqliteProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window, self)
        self.storage = Storage(window)
        self.id = "db_sqlite"
        self.type = "assistant_store"

    def attach(self, window):
        self.window = window
        self.storage.attach(window)

    def patch(self, version: Version) -> bool:
        """
        Patch versions

        :param version: current app version
        :return: True if migrated
        """
        return self.patcher.execute(version)

    def create_id(self) -> str:
        """
        Create unique uuid

        :return: uuid
        """
        return str(uuid.uuid4())

    def create(self, store: AssistantStoreItem) -> int:
        """
        Create new and return its ID

        :param store: AssistantStoreItem
        :return: store ID
        """
        if store.record_id is None or store.record_id == "":
            store.uuid = self.create_id()
            store.record_id = self.storage.insert(store)
        return store.record_id

    def load_all(self) -> dict:
        """
        Load stores from DB

        :return: stores dict
        """
        return self.storage.get_all()

    def load(self, id: int) -> AssistantStoreItem:
        """
        Load store from DB

        :param id: store ID
        :return: store item
        """
        return self.storage.get_by_id(id)

    def save(self, store: AssistantStoreItem):
        """
        Save store to DB

        :param store: AssistantStoreItem
        """
        try:
            self.storage.save(store)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving stored: {}".format(str(e)))

    def save_all(self, items: dict):
        """
        Save all stores to DB

        :param items: dict of AssistantStoreItem objects
        """
        try:
            for id in items:
                store = items[id]
                self.storage.save(store)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving store: {}".format(str(e)))

    def delete_by_id(self, id: int) -> bool:
        """
        Delete store item by IDx

        :param id: store item ID
        :return: True if deleted
        """
        return self.storage.delete_by_id(id)

    def delete_by_store_id(self, id: str) -> bool:
        """
        Delete store item by IDx

        :param id: store_id
        :return: True if deleted
        """
        return self.storage.delete_by_store_id(id)

    def truncate(self) -> bool:
        """
        Truncate all stores

        :return: True if truncated
        """
        return self.storage.truncate_all()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import BaseEvent


class ControlEvent(BaseEvent):
    """Events used for app control"""
    APP_STATUS = "app.status"
    APP_EXIT = "app.exit"
    CMD_CONFIRM = "cmd.confirm"
    AUDIO_OUTPUT_ENABLE = "audio.output.enable"
    AUDIO_OUTPUT_DISABLE = "audio.output.disable"
    AUDIO_INPUT_ENABLE = "audio.input.enable"
    AUDIO_INPUT_DISABLE = "audio.input.disable"
    CALENDAR_ADD = "calendar.add"
    CALENDAR_CLEAR = "calendar.clear"
    CALENDAR_READ = "calendar.read"
    CAMERA_ENABLE = "camera.enable"
    CAMERA_DISABLE = "camera.disable"
    CAMERA_CAPTURE = "camera.capture"
    CTX_NEW = "ctx.new"
    CTX_PREV = "ctx.prev"
    CTX_NEXT = "ctx.next"
    CTX_LAST = "ctx.last"
    CTX_INPUT_FOCUS = "ctx.input.focus"
    CTX_INPUT_SEND = "ctx.input.send"
    CTX_INPUT_CLEAR = "ctx.input.clear"
    CTX_STOP = "ctx.stop"
    CTX_ATTACHMENTS_CLEAR = "ctx.attachments.clear"
    CTX_CURRENT = "ctx.current"
    CTX_READ_LAST = "ctx.read.last"
    CTX_READ_ALL = "ctx.read.all"
    CTX_RENAME = "ctx.rename"
    CTX_SEARCH_STRING = "ctx.search.string"
    CTX_SEARCH_CLEAR = "ctx.search.clear"
    INPUT_SEND = "input.send"
    INPUT_APPEND = "input.append"
    MODE_CHAT = "mode.chat"
    MODE_LLAMA_INDEX = "mode.llama_index"
    NOTE_ADD = "note.add"
    NOTEPAD_CLEAR = "notepad.clear"
    NOTEPAD_READ = "notepad.read"
    TAB_CHAT = "tab.chat"
    TAB_CALENDAR = "tab.calendar"
    TAB_DRAW = "tab.draw"
    TAB_FILES = "tab.files"
    TAB_NOTEPAD = "tab.notepad"
    TAB_NEXT = "tab.next"
    TAB_PREV = "tab.prev"
    VOICE_MESSAGE_START = "voice_msg.start"
    VOICE_MESSAGE_STOP = "voice_msg.stop"
    VOICE_MESSAGE_TOGGLE = "voice_msg.toggle"
    VOICE_COMMAND_START = "voice_cmd.start"
    VOICE_COMMAND_STOP = "voice_cmd.stop"
    VOICE_COMMAND_TOGGLE = "voice_cmd.toggle"


class AppEvent(BaseEvent):
    """Events dispatched by application"""
    APP_STARTED = "app.started"
    CTX_CREATED = "ctx.created"
    CTX_END = "ctx.end"
    CTX_SELECTED = "ctx.selected"
    CTX_ATTACHMENTS_CLEAR = "ctx.attachments.clear"
    CAMERA_ENABLED = "camera.enabled"
    CAMERA_DISABLED = "camera.disabled"
    CAMERA_CAPTURED = "camera.captured"
    INPUT_ERROR = "input.error"
    INPUT_SENT = "input.sent"
    INPUT_CALL = "input.call"
    INPUT_STOPPED = "input.stopped"
    INPUT_VOICE_LISTEN_STARTED = "input.voice.listen.started"
    INPUT_VOICE_LISTEN_STOPPED = "input.voice.listen.stopped"
    VOICE_CONTROL_TOGGLE = "voice.control.toggle"
    VOICE_CONTROL_STARTED = "voice.control.started"
    VOICE_CONTROL_STOPPED = "voice.control.stopped"
    VOICE_CONTROL_SENT = "voice.control.sent"
    VOICE_CONTROL_UNRECOGNIZED = "voice.control.unrecognized"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from .common import Common
from .experts import Experts
from .flow import Flow


class Agent:
    def __init__(self, window=None):
        """
        Agent controller

        :param window: Window instance
        """
        self.window = window
        self.common = Common(window)
        self.experts = Experts(window)
        self.flow = Flow(window)
        self.options = {
            "agent.iterations": {
                "type": "int",
                "slider": True,
                "label": "agent.iterations",
                "min": 0,
                "max": 100,
                "step": 1,
                "value": 3,
                "multiplier": 1,
            },
        }

    def setup(self):
        """Setup agent controller"""
        # register hooks
        self.window.ui.add_hook("update.global.agent.iterations", self.hook_update)

        # restore config
        self.reload()

    def update(self):
        """Update agent status"""
        iterations = "-"
        mode = self.window.core.config.get('mode')

        # get iterations from plugin or from agent mode
        if mode == "agent":
            iterations = int(self.window.core.config.get("agent.iterations"))
        elif self.is_inline():
            if self.window.controller.plugins.is_enabled("agent"):
                iterations = int(self.window.core.plugins.get_option("agent", "iterations"))
        if iterations == 0:
            iterations_str = "∞"  # infinity loop
        else:
            iterations_str = str(iterations)

        status = str(self.flow.iteration) + " / " + iterations_str
        self.window.ui.nodes['status.agent'].setText(status)
        self.common.toggle_status()

    def reload(self):
        """Reload agent controller"""
        # auto-stop
        if self.window.core.config.get('agent.auto_stop'):
            self.window.ui.config['global']['agent.auto_stop'].setChecked(True)
        else:
            self.window.ui.config['global']['agent.auto_stop'].setChecked(False)

        # iterations
        self.window.controller.config.apply_value(
            parent_id="global",
            key="agent.iterations",
            option=self.options["agent.iterations"],
            value=self.window.core.config.get('agent.iterations'),
        )

    def hook_update(self, key, value, caller, *args, **kwargs):
        """
        Hook: on option update

        :param key: config key
        :param value: config value
        :param caller: caller name
        :param args: args
        :param kwargs: kwargs
        """
        if self.window.core.config.get(key) == value:
            return

        if key == 'agent.iterations':
            self.window.core.config.set(key, int(value))  # cast to int, if from text input
            self.window.core.config.save()
            self.update()

    def is_inline(self) -> bool:
        """
        Is agent inline (plugin) enabled

        :return: True if enabled
        """
        return self.window.controller.plugins.is_type_enabled("agent")

    def enabled(self) -> bool:
        """
        Is agent enabled

        :return: True if enabled
        """
        return self.window.core.config.get('mode') == 'agent' or self.is_inline()

    def add_run(self):
        """Increment agent iteration"""
        self.flow.iteration += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Mode:
    def __init__(self, window=None):
        """
        Mode controller

        :param window: Window instance
        """
        self.window = window
        self.locked = False

    def select(self, idx: int):
        """
        Select mode by idx

        :param idx: value of the list (row idx)
        """
        # check if mode change is not locked
        if self.change_locked():
            return
        mode = self.window.core.modes.get_by_idx(idx)
        self.set(mode)

        event = Event(Event.MODE_SELECT, {
            'value': mode,
        })
        self.window.core.dispatcher.dispatch(event)
        self.window.controller.attachment.update()

    def set(self, mode: str):
        """
        Set mode

        :param mode: mode name
        """
        self.locked = True
        # if ctx loaded with assistant ID assigned then switch to assistant from ctx
        if mode == "assistant":
            self.window.controller.presets.select_default()
            if self.window.core.ctx.current is not None \
                    and self.window.core.ctx.assistant is not None:
                self.window.controller.assistant.select_by_id(
                    self.window.core.ctx.assistant
                )
            else:
                self.window.controller.assistant.select_current()

        self.window.core.config.set('mode', mode)

        # reset model and preset at start
        self.window.core.config.set('model', "")
        self.window.core.config.set('preset', "")

        # update
        self.window.controller.attachment.update()
        self.window.controller.ctx.update_ctx()

        # update toolbox, mode, presets, model, assistant and rest of the UI
        self.window.controller.ui.update()

        # set status: ready
        self.window.ui.status(trans('status.started'))

        # if assistant mode then update ctx label
        if mode == "assistant":
            self.window.controller.ctx.common.update_label_by_current()
        self.locked = False

    def select_current(self):
        """Select current mode on the list"""
        mode = self.window.core.config.get('mode')
        if mode:
            idx = self.window.core.modes.get_idx_by_name(mode)
            current = self.window.ui.models['prompt.mode'].index(idx, 0)
            self.window.ui.nodes['prompt.mode'].setCurrentIndex(current)

    def select_default(self):
        """Set default mode"""
        mode = self.window.core.config.get('mode')
        if mode is None or mode == "":
            self.window.core.config.set('mode', self.window.core.modes.get_default())

    def default_all(self):
        """Set default mode, model and preset"""
        self.select_default()
        self.window.controller.model.select_default()
        self.window.controller.presets.select_default()
        self.window.controller.assistant.select_default()

    def update_list(self):
        """Update modes list"""
        self.window.ui.toolbox.mode.update(self.window.core.modes.get_all())

    def update_temperature(self, temperature: float = None):
        """
        Update current temperature field

        :param temperature: current temperature
        :type temperature: float or None
        """
        if temperature is None:
            if self.window.core.config.get('preset') is None or self.window.core.config.get('preset') == "":
                temperature = 1.0  # default temperature
            else:
                id = self.window.core.config.get('preset')
                if id in self.window.core.presets.items:
                    temperature = float(self.window.core.presets.items[id].temperature or 1.0)
        option = self.window.controller.settings.editor.get_options()["temperature"]
        '''
        self.window.controller.config.slider.on_update("global", "current_temperature", option, temperature,
                                                       hooks=False)  # disable hooks to prevent circular update
        '''

    def hook_global_temperature(self, key, value, caller, *args, **kwargs):
        """
        Hook: on update current temperature global field
        """
        if caller != "slider":
            return  # accept call only from slider (has already validated min/max)

        temperature = value / 100
        self.window.core.config.set("temperature", temperature)
        preset_id = self.window.core.config.get('preset')
        if preset_id is not None and preset_id != "":
            if preset_id in self.window.core.presets.items:
                preset = self.window.core.presets.items[preset_id]
                preset.temperature = temperature
                self.window.core.presets.save(preset_id)

    def switch_inline(self, mode: str, ctx: CtxItem, prompt: str) -> str:
        """
        Switch inline mode

        :param mode: current mode
        :param ctx: ctx item
        :param prompt: prompt text
        :return: changed mode
        """
        event = Event(Event.MODE_BEFORE, {
            'value': mode,
            'prompt': prompt,
        })
        event.ctx = ctx
        self.window.core.dispatcher.dispatch(event)
        return event.data['value']

    def update_mode(self):
        """Update mode"""
        self.select_default()  # set default mode
        self.update_list()  # update modes list
        self.select_current()  # select current mode on the list
        self.window.controller.chat.vision.update()

    def reset_current(self):
        """Reset current setup"""
        self.window.core.config.set('prompt', None)
        self.window.core.config.set('ai_name', None)
        self.window.core.config.set('user_name', None)

    def change_locked(self) -> bool:
        """
        Check if mode change is locked

        :return: True if locked
        :rtype: bool
        """
        return self.window.controller.chat.input.generating



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.12 10:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon, QImage
from PySide6.QtWidgets import QTextEdit, QApplication

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ChatInput(QTextEdit):
    def __init__(self, window=None):
        """
        Chat input

        :param window: main window
        """
        super(ChatInput, self).__init__(window)
        self.window = window
        self.setAcceptRichText(False)
        self.setFocus()
        self.value = self.window.core.config.data['font_size.input']
        self.max_font_size = 42
        self.min_font_size = 8
        self.textChanged.connect(self.window.controller.ui.update_tokens)

    def insertFromMimeData(self, source):
        """
        Insert from mime data

        :param source: source
        """
        self.handle_clipboard(source)
        if not source.hasImage():
            super().insertFromMimeData(source)

    def handle_clipboard(self, source):
        """
        Handle clipboard

        :param source: source
        """
        if source.hasImage():
            image = source.imageData()
            if isinstance(image, QImage):
                self.window.controller.attachment.from_clipboard_image(image)
        elif source.hasUrls():
            urls = source.urls()
            for url in urls:
                if url.isLocalFile():
                    local_path = url.toLocalFile()
                    self.window.controller.attachment.from_clipboard_url(local_path)
        elif source.hasText():
            text = source.text()
            self.window.controller.attachment.from_clipboard_text(text)

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: event
        """
        menu = self.createStandardContextMenu()

        # paste attachment
        if self.window.controller.attachment.clipboard_has_attachment():
            action = QAction(QIcon(":/icons/paste.svg"), trans("action.use.attachment"), self)
            action.triggered.connect(self.action_from_clipboard)
            menu.addAction(action)

        selected_text = self.textCursor().selectedText()
        if selected_text:
            # plain text
            plain_text = self.textCursor().selection().toPlainText()

            # audio read
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(self, selected_text, excluded=["input"])
            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text))
            menu.addAction(action)
        else:
            # save as (all)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText()))
            menu.addAction(action)

        self.window.core.prompt.template.to_menu_options(menu, "input")

        menu.exec_(event.globalPos())

    def action_from_clipboard(self):
        """
        Get from clipboard
        """
        clipboard = QApplication.clipboard()
        source = clipboard.mimeData()
        self.handle_clipboard(source)

    def audio_read_selection(self):
        """Read selected text (audio)"""
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        super(ChatInput, self).keyPressEvent(event)
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            mode = self.window.core.config.get('send_mode')
            if mode > 0:  # Enter or Shift + Enter
                if mode == 2:  # Shift + Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers == QtCore.Qt.ShiftModifier or modifiers == QtCore.Qt.ControlModifier:
                        self.window.controller.chat.input.send_input()
                else:  # Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers != QtCore.Qt.ShiftModifier and modifiers != QtCore.Qt.ControlModifier:
                        self.window.controller.chat.input.send_input()
                self.setFocus()

        # cancel edit
        elif event.key() == Qt.Key_Escape and self.window.controller.ctx.extra.is_editing():
            self.window.controller.ctx.extra.edit_cancel()

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            self.window.core.config.data['font_size.input'] = self.value
            self.window.core.config.save()
            self.window.controller.ui.update_font_size()
            event.accept()
        else:
            super(ChatInput, self).wheelEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import copy

from PySide6 import QtCore
from PySide6.QtCore import QObject, QEvent

from pygpt_net.core.access.events import ControlEvent


class Shortcuts:
    def __init__(self, window=None):
        """
        Shortcuts core

        :param window: Window instance
        """
        self.window = window

    def get_keys_choices(self) -> list:
        """
        Get keys choices with identical key and value.
        """
        keys = [{chr(i): chr(i)} for i in range(65, 91)]  # A-Z
        keys.extend([{str(i): str(i)} for i in range(0, 10)])  # 0-9
        keys.extend([{str(i): str(i)} for i in range(1, 13)])  # F1-F12
        symbols = [
            'Space', 'Comma', 'Period', 'Slash', 'Backslash', 'Minus', 'Equal',
            'Semicolon', 'Apostrophe'
        ]
        keys.extend([{name: name} for name in symbols])
        return keys

    def get_modifiers_choices(self) -> list:
        """
        Get modifiers choices with identical key and value.
        """
        modifiers_names = [
            'Shift', 'Control', 'Meta', 'Alt', 'Escape', 'Tab', 'Backtab',
            'Backspace', 'Return', 'Enter', 'Insert', 'Delete', 'Pause',
            'Print', 'SysReq', 'Clear', 'Home', 'End', 'Left', 'Up',
            'Right', 'Down', 'PageUp', 'PageDown', 'Super_L', 'Super_R', 'Menu'
        ]
        modifiers = [{name: name} for name in modifiers_names]
        return modifiers


class GlobalShortcutFilter(QObject):
    def __init__(self, window=None):
        super().__init__()
        self.window = window

    def eventFilter(self, obj, event):
        """
        Event filter for global shortcuts handling

        :param obj: object
        :param event: event
        :return: True if event was handled
        """
        try:
            # esc key
            if event.type() == QEvent.KeyPress and event.key() == QtCore.Qt.Key_Escape:
                self.window.controller.access.on_escape()

            # shortcuts
            config = copy.deepcopy(self.window.core.config.get("access.shortcuts"))
            for shortcut in config:
                if shortcut['key'] == "" or shortcut['key'] is None:
                    continue
                shortcut['key'] = getattr(QtCore.Qt, 'Key_' + str(shortcut['key']))
                shortcut['key_modifier'] = getattr(QtCore.Qt, str(shortcut['key_modifier']) + 'Modifier', QtCore.Qt.NoModifier)
            if event.type() == QEvent.KeyPress:
                for shortcut in config:
                    if (event.key() == shortcut['key'] and
                        (shortcut['key_modifier'] == QtCore.Qt.NoModifier or event.modifiers() == shortcut['key_modifier'])
                    ):
                        self.window.core.dispatcher.dispatch(ControlEvent(shortcut['action']))
                        return True
        except Exception as e:
            print(e)
            pass
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import copy
import os

from packaging.version import parse as parse_version, Version


class Patch:
    def __init__(self, window=None):
        self.window = window

    def execute(self, version: Version) -> bool:
        """
        Migrate config to current app version

        :param version: current app version
        :return: True if migrated
        """
        data = self.window.core.config.all()
        current = "0.0.0"
        updated = False
        is_old = False

        # get version of config file
        if '__meta__' in data and 'version' in data['__meta__']:
            current = data['__meta__']['version']
        old = parse_version(current)

        # check if config file is older than current app version
        if old < version:

            # mark as older version
            is_old = True

            # < 0.9.1
            if old < parse_version("0.9.1"):
                print("Migrating config from < 0.9.1...")
                keys_to_remove = ['user_id', 'custom']  # not needed anymore
                for key in keys_to_remove:
                    if key in data:
                        del data[key]
                keys_to_add = ['organization_key']
                for key in keys_to_add:
                    if key not in data:
                        data[key] = ""
                updated = True

            # < 0.9.2
            if old < parse_version("0.9.2"):
                print("Migrating config from < 0.9.2...")
                keys_to_remove = ['ui.ctx.min_width',
                                  'ui.ctx.max_width',
                                  'ui.toolbox.min_width',
                                  'ui.toolbox.max_width',
                                  'ui.dialog.settings.width',
                                  'ui.dialog.settings.height',
                                  'ui.chatbox.font.color']
                for key in keys_to_remove:
                    if key in data:
                        del data[key]
                if 'theme' not in data:
                    data['theme'] = "dark_teal"
                updated = True

            # < 0.9.4
            if old < parse_version("0.9.4"):
                print("Migrating config from < 0.9.4...")
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'plugins_enabled' not in data:
                    data['plugins_enabled'] = {}
                updated = True

            # < 0.9.6
            if old < parse_version("0.9.6"):
                print("Migrating config from < 0.9.6...")
                data['debug'] = True  # enable debug by default
                updated = True

            # < 2.0.0
            if old < parse_version("2.0.0"):
                print("Migrating config from < 2.0.0...")
                data['theme'] = 'dark_teal'  # force, because removed light themes!
                if 'cmd' not in data:
                    data['cmd'] = True
                if 'stream' not in data:
                    data['stream'] = True
                if 'attachments_send_clear' not in data:
                    data['attachments_send_clear'] = True
                if 'assistant' not in data:
                    data['assistant'] = None
                if 'assistant_thread' not in data:
                    data['assistant_thread'] = None
                updated = True

            # < 2.0.1
            if old < parse_version("2.0.1"):
                print("Migrating config from < 2.0.1...")
                if 'send_mode' not in data:
                    data['send_mode'] = 1
                if 'send_shift_enter' in data:
                    del data['send_shift_enter']
                if 'font_size.input' not in data:
                    data['font_size.input'] = 11
                if 'font_size.ctx' not in data:
                    data['font_size.ctx'] = 9
                if 'ctx.auto_summary' not in data:
                    data['ctx.auto_summary'] = True
                if 'ctx.auto_summary.system' not in data:
                    data['ctx.auto_summary.system'] = "You are an expert in conversation summarization"
                if 'ctx.auto_summary.prompt' not in data:
                    data['ctx.auto_summary.prompt'] = "Summarize topic of this conversation in one sentence. Use best " \
                                                      "keywords to describe it. Summary must be in the same language " \
                                                      "as the conversation and it will be used for conversation title " \
                                                      "so it must be EXTREMELY SHORT and concise - use maximum 5 " \
                                                      "words: \n\nUser: {input}\nAI Assistant: {output}"
                updated = True

            # < 2.0.6
            if old < parse_version("2.0.6"):
                print("Migrating config from < 2.0.6...")
                if 'layout.density' not in data:
                    data['layout.density'] = -2
                updated = True

            # < 2.0.8
            if old < parse_version("2.0.8"):
                print("Migrating config from < 2.0.8...")
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'cmd_web_google' not in data['plugins']:
                    data['plugins']['cmd_web_google'] = {}
                data['plugins']['cmd_web_google'][
                    'prompt_summarize'] = "Summarize the English text in a maximum of 3 " \
                                          "paragraphs, trying to find the most " \
                                          "important content that can help answer the " \
                                          "following question: "
                data['plugins']['cmd_web_google']['chunk_size'] = 100000
                data['plugins']['cmd_web_google']['max_page_content_length'] = 0
                updated = True

            # < 2.0.13
            if old < parse_version("2.0.13"):
                print("Migrating config from < 2.0.13...")
                if 'layout.density' not in data:
                    data['layout.density'] = 0
                else:
                    if data['layout.density'] == -2:
                        data['layout.density'] = 0
                updated = True

            # < 2.0.14
            if old < parse_version("2.0.14"):
                print("Migrating config from < 2.0.14...")
                if 'vision.capture.enabled' not in data:
                    data['vision.capture.enabled'] = True
                if 'vision.capture.auto' not in data:
                    data['vision.capture.auto'] = True
                if 'vision.capture.width' not in data:
                    data['vision.capture.width'] = 800
                if 'vision.capture.height' not in data:
                    data['vision.capture.height'] = 600
                updated = True

            # < 2.0.16
            if old < parse_version("2.0.16"):
                print("Migrating config from < 2.0.16...")
                if 'vision.capture.idx' not in data:
                    data['vision.capture.idx'] = 0
                if 'img_raw' not in data:
                    data['img_raw'] = True
                if 'img_prompt_model' not in data:
                    data['img_prompt_model'] = "gpt-4-1106-preview"
                updated = True

            # < 2.0.19
            if old < parse_version("2.0.19"):
                print("Migrating config from < 2.0.19...")
                if 'img_raw' not in data:
                    data['img_raw'] = True
                if not data['img_raw']:
                    data['img_raw'] = True
                updated = True

            # < 2.0.25
            if old < parse_version("2.0.25"):
                print("Migrating config from < 2.0.25...")
                if 'cmd.prompt' not in data:
                    data['cmd.prompt'] = self.window.core.config.get_base('cmd.prompt')
                if 'img_prompt' not in data:
                    data['img_prompt'] = self.window.core.config.get_base('img_prompt')
                if 'vision.capture.quality' not in data:
                    data['vision.capture.quality'] = 85
                if 'attachments_capture_clear' not in data:
                    data['attachments_capture_clear'] = True
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'cmd_web_google' not in data['plugins']:
                    data['plugins']['cmd_web_google'] = {}
                data['plugins']['cmd_web_google']['prompt_summarize'] = "Summarize the English text in a maximum of 3 " \
                                                                        "paragraphs, trying to find the most " \
                                                                        "important content that can help answer the " \
                                                                        "following question: "
                updated = True

            # < 2.0.26
            if old < parse_version("2.0.26"):
                print("Migrating config from < 2.0.26...")
                if 'ctx.auto_summary.model' not in data:
                    data['ctx.auto_summary.model'] = 'gpt-3.5-turbo-1106'
                updated = True

            # < 2.0.27
            if old < parse_version("2.0.27"):
                print("Migrating config from < 2.0.27...")
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'cmd_web_google' not in data['plugins']:
                    data['plugins']['cmd_web_google'] = {}
                data['plugins']['cmd_web_google'][
                    'prompt_summarize'] = "Summarize text in English in a maximum of 3 " \
                                          "paragraphs, trying to find the most " \
                                          "important content that can help answer the " \
                                          "following question: {query}"
                data['cmd.prompt'] = self.window.core.config.get_base('cmd.prompt')  # fix
                updated = True

            # < 2.0.30
            if old < parse_version("2.0.30"):
                print("Migrating config from < 2.0.30...")
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'audio_openai_whisper' not in data['plugins']:
                    data['plugins']['audio_openai_whisper'] = {}
                data['plugins']['audio_openai_whisper']['timeout'] = 1
                data['plugins']['audio_openai_whisper']['phrase_length'] = 5
                data['plugins']['audio_openai_whisper']['min_energy'] = 2000
                updated = True

            # < 2.0.31
            if old < parse_version("2.0.31"):
                print("Migrating config from < 2.0.31...")
                if 'plugins' not in data:
                    data['plugins'] = {}
                if 'audio_openai_whisper' not in data['plugins']:
                    data['plugins']['audio_openai_whisper'] = {}
                data['plugins']['audio_openai_whisper']['continuous_listen'] = False
                data['plugins']['audio_openai_whisper']['timeout'] = 2
                data['plugins']['audio_openai_whisper']['phrase_length'] = 4
                data['plugins']['audio_openai_whisper']['magic_word_timeout'] = 1
                data['plugins']['audio_openai_whisper']['magic_word_phrase_length'] = 2
                data['plugins']['audio_openai_whisper']['min_energy'] = 1.3
                updated = True

            # < 2.0.34
            if old < parse_version("2.0.34"):
                print("Migrating config from < 2.0.34...")
                if 'lock_modes' not in data:
                    data['lock_modes'] = True
                updated = True

            # < 2.0.37
            if old < parse_version("2.0.37"):
                print("Migrating config from < 2.0.37...")
                if 'font_size.toolbox' not in data:
                    data['font_size.toolbox'] = 12
                updated = True

            # < 2.0.46
            if old < parse_version("2.0.46"):
                print("Migrating config from < 2.0.46...")
                data['cmd'] = False  # disable on default
                updated = True

            # < 2.0.47
            if old < parse_version("2.0.47"):
                print("Migrating config from < 2.0.47...")
                if 'notepad.num' not in data:
                    data['notepad.num'] = 5
                updated = True

            # < 2.0.52
            if old < parse_version("2.0.52"):
                print("Migrating config from < 2.0.52...")
                if 'layout.dpi.scaling' not in data:
                    data['layout.dpi.scaling'] = True
                if 'layout.dpi.factor' not in data:
                    data['layout.dpi.factor'] = 1.0
                updated = True

            # < 2.0.62
            if old < parse_version("2.0.62"):
                print("Migrating config from < 2.0.62...")
                if 'ctx.records.limit' not in data:
                    data['ctx.records.limit'] = 0
                updated = True

            # < 2.0.65
            if old < parse_version("2.0.65"):
                print("Migrating config from < 2.0.65...")
                if 'ctx.search.string' not in data:
                    data['ctx.search.string'] = ""
                updated = True

            # < 2.0.69
            if old < parse_version("2.0.69"):
                print("Migrating config from < 2.0.69...")
                data['img_prompt'] = "Whenever I provide a basic idea or concept for an image, such as 'a picture " \
                                     "of mountains', I want you to ALWAYS translate it into English and expand " \
                                     "and elaborate on this idea. Use your knowledge and creativity to add " \
                                     "details that would make the image more vivid and interesting. This could " \
                                     "include specifying the time of day, weather conditions, surrounding " \
                                     "environment, and any additional elements that could enhance the scene. Your " \
                                     "goal is to create a detailed and descriptive prompt that provides DALL-E " \
                                     "with enough information to generate a rich and visually appealing image. " \
                                     "Remember to maintain the original intent of my request while enriching the " \
                                     "description with your imaginative details."
                data['img_raw'] = False
                data['img_prompt_model'] = "gpt-4-1106-preview"
                updated = True

            # < 2.0.71
            if old < parse_version("2.0.71"):
                print("Migrating config from < 2.0.71...")
                prompt = 'IMAGE GENERATION: Whenever I provide a basic idea or concept for an image, such as \'a picture of ' \
                         'mountains\', I want you to ALWAYS translate it into English and expand and elaborate on this idea. ' \
                         'Use your  knowledge and creativity to add details that would make the image more vivid and ' \
                         'interesting. This could include specifying the time of day, weather conditions, surrounding ' \
                         'environment, and any additional elements that could enhance the scene. Your goal is to create a ' \
                         'detailed and descriptive prompt that provides DALL-E  with enough information to generate a rich ' \
                         'and visually appealing image. Remember to maintain the original  intent of my request while ' \
                         'enriching the description with your imaginative details. HOW TO START IMAGE GENERATION: to start ' \
                         'image generation return to me prepared prompt in JSON format, all in one line,  using following ' \
                         'syntax: ~###~{"cmd": "image", "params": {"query": "your query here"}}~###~. Use ONLY this syntax ' \
                         'and remember to surround JSON string with ~###~.  DO NOT use any other syntax. Use English in the ' \
                         'generated JSON command, but conduct all the remaining parts of the discussion with me in the ' \
                         'language in which I am speaking to you. The image will be generated on my machine  immediately ' \
                         'after the command is issued, allowing us to discuss the photo once it has been created.  Please ' \
                         'engage with me about the photo itself, not only by giving the generate command. '
                if 'openai_dalle' not in data['plugins']:
                    data['plugins']['openai_dalle'] = {}
                    data['plugins']['openai_dalle']['prompt'] = prompt  # fixed prompt

                data['plugins_enabled']['openai_dalle'] = True
                data['plugins_enabled']['openai_vision'] = True

                # deprecate vision and img modes
                if data['mode'] == 'vision' or data['mode'] == 'img':
                    data['mode'] = 'chat'

                updated = True

            # < 2.0.72
            if old < parse_version("2.0.72"):
                print("Migrating config from < 2.0.72...")
                if 'theme.markdown' not in data:
                    data['theme.markdown'] = True
                prompt = 'IMAGE GENERATION: Whenever I provide a basic idea or concept for an image, such as \'a picture of ' \
                         'mountains\', I want you to ALWAYS translate it into English and expand and elaborate on this idea. ' \
                         'Use your  knowledge and creativity to add details that would make the image more vivid and ' \
                         'interesting. This could include specifying the time of day, weather conditions, surrounding ' \
                         'environment, and any additional elements that could enhance the scene. Your goal is to create a ' \
                         'detailed and descriptive prompt that provides DALL-E  with enough information to generate a rich ' \
                         'and visually appealing image. Remember to maintain the original  intent of my request while ' \
                         'enriching the description with your imaginative details. HOW TO START IMAGE GENERATION: to start ' \
                         'image generation return to me prepared prompt in JSON format, all in one line,  using following ' \
                         'syntax: ~###~{"cmd": "image", "params": {"query": "your query here"}}~###~. Use ONLY this syntax ' \
                         'and remember to surround JSON string with ~###~.  DO NOT use any other syntax. Use English in the ' \
                         'generated JSON command, but conduct all the remaining parts of the discussion with me in the ' \
                         'language in which I am speaking to you. The image will be generated on my machine  immediately ' \
                         'after the command is issued, allowing us to discuss the photo once it has been created.  Please ' \
                         'engage with me about the photo itself, not only by giving the generate command. '
                if 'openai_dalle' not in data['plugins']:
                    data['plugins']['openai_dalle'] = {}
                data['plugins']['openai_dalle']['prompt'] = prompt  # fixed prompt
                updated = True

            # < 2.0.75
            if old < parse_version("2.0.75"):
                print("Migrating config from < 2.0.75...")
                if 'updater.check.launch' not in data:
                    data['updater.check.launch'] = True
                if 'updater.check.bg' not in data:
                    data['updater.check.bg'] = False
                updated = True

            # < 2.0.78
            if old < parse_version("2.0.78"):
                print("Migrating config from < 2.0.78...")
                if 'render.plain' not in data:
                    data['render.plain'] = False
                updated = True

            # < 2.0.81
            if old < parse_version("2.0.81"):
                print("Migrating config from < 2.0.81...")
                self.window.core.updater.patch_css('markdown.light.css', True)  # force replace file
                updated = True

            # < 2.0.85
            if old < parse_version("2.0.85"):
                print("Migrating config from < 2.0.85...")
                prompt = "AUTONOMOUS MODE:\n1. You will now enter self-dialogue mode, where you will be conversing with " \
                         "yourself, not with a human.\n2. When you enter self-dialogue mode, remember that you are engaging " \
                         "in a conversation with yourself. Any user input will be considered a reply featuring your previous response.\n" \
                         "3. The objective of this self-conversation is well-defined—focus on achieving it.\n" \
                         "4. Your new message should be a continuation of the last response you generated, essentially replying" \
                         " to yourself and extending it.\n5. After each response, critically evaluate its effectiveness " \
                         "and alignment with the goal. If necessary, refine your approach.\n6. Incorporate self-critique " \
                         "after every response to capitalize on your strengths and address areas needing improvement.\n7. To " \
                         "advance towards the goal, utilize all the strategic thinking and resources at your disposal.\n" \
                         "8. Ensure that the dialogue remains coherent and logical, with each response serving as a stepping " \
                         "stone towards the ultimate objective.\n9. Treat the entire dialogue as a monologue aimed at devising" \
                         " the best possible solution to the problem.\n10. Conclude the self-dialogue upon realizing the " \
                         "goal or reaching a pivotal conclusion that meets the initial criteria.\n11. You are allowed to use " \
                         "any commands and tools without asking for it.\n12. While using commands, always use the correct " \
                         "syntax and never interrupt the command before generating the full instruction.\n13. ALWAYS break " \
                         "down the main task into manageable logical subtasks, systematically addressing and analyzing each" \
                         " one in sequence.\n14. With each subsequent response, make an effort to enhance your previous " \
                         "reply by enriching it with new ideas and do it automatically without asking for it.\n14. Any input " \
                         "that begins with 'user: ' will come from me, and I will be able to provide you with ANY additional " \
                         "commands or goal updates in this manner. The other inputs, not prefixed with 'user: ' will represent" \
                         " your previous responses.\n15. Start by breaking down the task into as many smaller sub-tasks as " \
                         "possible, then proceed to complete each one in sequence.  Next, break down each sub-task into even " \
                         "smaller tasks, carefully and step by step go through all of them until the required goal is fully " \
                         "and correctly achieved.\n"
                if 'self_loop' not in data['plugins']:
                    data['plugins']['self_loop'] = {}
                data['plugins']['self_loop']['prompt'] = prompt  # fixed prompt

                # before fix (from 2.0.72)
                prompt = 'IMAGE GENERATION: Whenever I provide a basic idea or concept for an image, such as \'a picture of ' \
                         'mountains\', I want you to ALWAYS translate it into English and expand and elaborate on this idea. ' \
                         'Use your  knowledge and creativity to add details that would make the image more vivid and ' \
                         'interesting. This could include specifying the time of day, weather conditions, surrounding ' \
                         'environment, and any additional elements that could enhance the scene. Your goal is to create a ' \
                         'detailed and descriptive prompt that provides DALL-E  with enough information to generate a rich ' \
                         'and visually appealing image. Remember to maintain the original  intent of my request while ' \
                         'enriching the description with your imaginative details. HOW TO START IMAGE GENERATION: to start ' \
                         'image generation return to me prepared prompt in JSON format, all in one line,  using following ' \
                         'syntax: ~###~{"cmd": "image", "params": {"query": "your query here"}}~###~. Use ONLY this syntax ' \
                         'and remember to surround JSON string with ~###~.  DO NOT use any other syntax. Use English in the ' \
                         'generated JSON command, but conduct all the remaining parts of the discussion with me in the ' \
                         'language in which I am speaking to you. The image will be generated on my machine  immediately ' \
                         'after the command is issued, allowing us to discuss the photo once it has been created.  Please ' \
                         'engage with me about the photo itself, not only by giving the generate command. '
                if 'openai_dalle' not in data['plugins']:
                    data['plugins']['openai_dalle'] = {}
                data['plugins']['openai_dalle']['prompt'] = prompt  # fixed prompt

                updated = True

            # < 2.0.88
            if old < parse_version("2.0.88"):
                print("Migrating config from < 2.0.88...")
                prompt = "AUTONOMOUS MODE:\n1. You will now enter self-dialogue mode, where you will be conversing with " \
                         "yourself, not with a human.\n2. When you enter self-dialogue mode, remember that you are engaging " \
                         "in a conversation with yourself. Any user input will be considered a reply featuring your previous response.\n" \
                         "3. The objective of this self-conversation is well-defined—focus on achieving it.\n" \
                         "4. Your new message should be a continuation of the last response you generated, essentially replying" \
                         " to yourself and extending it.\n5. After each response, critically evaluate its effectiveness " \
                         "and alignment with the goal. If necessary, refine your approach.\n6. Incorporate self-critique " \
                         "after every response to capitalize on your strengths and address areas needing improvement.\n7. To " \
                         "advance towards the goal, utilize all the strategic thinking and resources at your disposal.\n" \
                         "8. Ensure that the dialogue remains coherent and logical, with each response serving as a stepping " \
                         "stone towards the ultimate objective.\n9. Treat the entire dialogue as a monologue aimed at devising" \
                         " the best possible solution to the problem.\n10. Conclude the self-dialogue upon realizing the " \
                         "goal or reaching a pivotal conclusion that meets the initial criteria.\n11. You are allowed to use " \
                         "any commands and tools without asking for it.\n12. While using commands, always use the correct " \
                         "syntax and never interrupt the command before generating the full instruction.\n13. ALWAYS break " \
                         "down the main task into manageable logical subtasks, systematically addressing and analyzing each" \
                         " one in sequence.\n14. With each subsequent response, make an effort to enhance your previous " \
                         "reply by enriching it with new ideas and do it automatically without asking for it.\n15. Any input " \
                         "that begins with 'user: ' will come from me, and I will be able to provide you with ANY additional " \
                         "commands or goal updates in this manner. The other inputs, not prefixed with 'user: ' will represent" \
                         " your previous responses.\n16. Start by breaking down the task into as many smaller sub-tasks as " \
                         "possible, then proceed to complete each one in sequence.  Next, break down each sub-task into even " \
                         "smaller tasks, carefully and step by step go through all of them until the required goal is fully " \
                         "and correctly achieved.\n"
                if 'self_loop' not in data['plugins']:
                    data['plugins']['self_loop'] = {}
                data['plugins']['self_loop']['prompt'] = prompt  # fixed prompt
                updated = True

            # < 2.0.91
            if old < parse_version("2.0.91"):
                print("Migrating config from < 2.0.91...")
                self.window.core.updater.patch_css('style.dark.css', True)  # force replace file
                updated = True

            # < 2.0.96
            if old < parse_version("2.0.96"):
                print("Migrating config from < 2.0.96...")
                if 'img_quality' not in data:
                    data['img_quality'] = "standard"
                updated = True

            # < 2.0.98
            if old < parse_version("2.0.98"):
                print("Migrating config from < 2.0.98...")
                data['img_resolution'] = "1792x1024"  # char fix
                self.window.core.updater.patch_css('style.css', True)  # force replace file
                self.window.core.updater.patch_css('style.light.css', True)  # force replace file
                self.window.core.updater.patch_css('style.dark.css', True)  # force replace file
                updated = True

            # < 2.0.99
            if old < parse_version("2.0.99"):
                print("Migrating config from < 2.0.99...")
                if 'layout.splitters' in data:
                    if 'calendar' in data['layout.splitters']:
                        # restore if was hidden at < 2.0.99
                        if len(data['layout.splitters']['calendar']) == 2 \
                                and data['layout.splitters']['calendar'][1] == 0:
                            data['layout.splitters']['calendar'] = [100, 100]
                if data['layout.density'] == 0:
                    data['layout.density'] = -1
                updated = True

            # < 2.0.100
            if old < parse_version("2.0.100"):
                print("Migrating config from < 2.0.100...")
                # rename output dir to data dir
                src = os.path.join(self.window.core.config.path, 'output')
                dst = os.path.join(self.window.core.config.path, 'data')

                # migrate data dir name
                if os.path.exists(src):
                    # backup old data dir
                    if os.path.exists(dst):
                        backup = os.path.join(self.window.core.config.path, 'data.backup')
                        os.rename(dst, backup)
                    # rename "output" to "data"
                    if os.path.exists(src):
                        os.rename(src, dst)

                # add llama-index config keys:
                if "llama.idx.auto" not in data:
                    data["llama.idx.auto"] = False
                if "llama.idx.auto.index" not in data:
                    data["llama.idx.auto.index"] = "base"
                if "llama.idx.current" not in data:
                    data["llama.idx.current"] = "base"
                if "llama.idx.db.index" not in data:
                    data["llama.idx.db.index"] = ""
                if "llama.idx.db.last" not in data:
                    data["llama.idx.db.last"] = 0
                if "llama.idx.list" not in data:
                    data["llama.idx.list"] = [
                        {
                            "id": "base",
                            "name": "Base",
                        }
                    ]
                if "llama.idx.status" not in data:
                    data["llama.idx.status"] = {}

                if "llama.log" not in data:
                    data["llama.log"] = False

                updated = True

            # < 2.0.101
            if old < parse_version("2.0.101"):
                print("Migrating config from < 2.0.101...")
                if 'layout.tooltips' not in data:
                    data['layout.tooltips'] = True
                updated = True

            # < 2.0.102
            if old < parse_version("2.0.102"):
                print("Migrating config from < 2.0.102...")
                if 'llama.hub.loaders' not in data:
                    data['llama.hub.loaders'] = [
                        {
                            "ext": "pptx",
                            "loader": "PptxReader"
                        },
                        {
                            "ext": "png,jpg,jpeg",
                            "loader": "ImageReader"
                        }
                    ]
                updated = True

            # < 2.0.112
            if old < parse_version("2.0.112"):
                print("Migrating config from < 2.0.112...")
                if 'img_dialog_open' not in data:
                    data['img_dialog_open'] = True
                updated = True

            # < 2.0.114
            if old < parse_version("2.0.114"):
                print("Migrating config from < 2.0.114...")
                if 'llama.idx.storage' not in data:
                    data['llama.idx.storage'] = "SimpleVectorStore"
                if 'llama.idx.storage.args' not in data:
                    data['llama.idx.storage.args'] = []
                if 'llama.idx.raw' not in data:
                    data['llama.idx.raw'] = False
                updated = True

            # < 2.0.116
            if old < parse_version("2.0.116"):
                print("Migrating config from < 2.0.116...")
                data['debug'] = False
                updated = True

            # < 2.0.118
            if old < parse_version("2.0.118"):
                print("Migrating config from < 2.0.118...")
                if 'layout.tray' not in data:
                    data['layout.tray'] = True
                updated = True

            """
            # < 2.0.119
            if old < parse_version("2.0.119"):
                print("Migrating config from < 2.0.119...")
                if 'layout.minimized' not in data:
                    data['layout.minimized'] = False
                updated = True
            """

            # < 2.0.121
            if old < parse_version("2.0.121"):
                print("Migrating config from < 2.0.121...")
                if 'openai_vision' not in data['plugins']:
                    data['plugins']['openai_vision'] = {}
                data['plugins']['openai_vision']['model'] = "gpt-4-vision-preview"
                updated = True

            # < 2.0.123
            if old < parse_version("2.0.123"):
                print("Migrating config from < 2.0.123...")
                if 'llama.idx.recursive' not in data:
                    data['llama.idx.recursive'] = False
                updated = True

            # < 2.0.127
            if old < parse_version("2.0.127"):
                print("Migrating config from < 2.0.127...")
                if 'upload.store' not in data:
                    data['upload.store'] = True
                if 'upload.data_dir' not in data:
                    data['upload.data_dir'] = False
                updated = True

            # < 2.0.131
            if old < parse_version("2.0.131"):
                print("Migrating config from < 2.0.131...")
                if 'self_loop' in data['plugins'] \
                        and 'prompts' not in data['plugins']['self_loop'] \
                        and 'prompt' in data['plugins']['self_loop'] \
                        and 'extended_prompt' in data['plugins']['self_loop']:

                    # copy old prompts to new list of prompts
                    data['plugins']['self_loop']['prompts'] = [
                        {
                            "enabled": True,
                            "name": "Default",
                            "prompt": data['plugins']['self_loop']['prompt'],
                        },
                        {
                            "enabled": False,
                            "name": "Extended",
                            "prompt": data['plugins']['self_loop']['extended_prompt'],
                        },
                    ]
                    updated = True

            # < 2.0.132
            if old < parse_version("2.0.132"):
                print("Migrating config from < 2.0.132...")
                if 'agent.auto_stop' not in data:
                    data['agent.auto_stop'] = True
                if 'agent.iterations' not in data:
                    data['agent.iterations'] = 3
                updated = True

            # < 2.0.135
            if old < parse_version("2.0.135"):
                print("Migrating config from < 2.0.135...")
                if 'agent.mode' not in data:
                    data['agent.mode'] = "chat"
                if 'agent.idx' not in data:
                    data['agent.idx'] = "base"
                updated = True

            # < 2.0.138
            if old < parse_version("2.0.138"):
                print("Migrating config from < 2.0.138...")
                if 'layout.tray.minimize' not in data:
                    data['layout.tray.minimize'] = False
                updated = True

            # < 2.0.139
            if old < parse_version("2.0.139"):
                print("Migrating config from < 2.0.139...")
                data['updater.check.bg'] = True
                if 'license.accepted' not in data:
                    data['license.accepted'] = False
                if 'updater.check.bg.last_time' not in data:
                    data['updater.check.bg.last_time'] = None
                if 'updater.check.bg.last_version' not in data:
                    data['updater.check.bg.last_version'] = None
                updated = True

            # < 2.0.142
            if old < parse_version("2.0.142"):
                print("Migrating config from < 2.0.142...")
                if 'agent.goal.notify' not in data:
                    data['agent.goal.notify'] = True
                updated = True

            # < 2.0.143
            if old < parse_version("2.0.143"):
                print("Migrating config from < 2.0.143...")
                if 'ctx.records.filter' not in data:
                    data['ctx.records.filter'] = "all"
                updated = True

            # < 2.0.144
            if old < parse_version("2.0.144"):
                print("Migrating config from < 2.0.144...")
                if 'cmd_history' in data['plugins'] \
                        and 'syntax_get_ctx_list_in_date_range' in data['plugins']['cmd_history']:
                    # remove
                    del data['plugins']['cmd_history']['syntax_get_ctx_list_in_date_range']
                if 'cmd_history' in data['plugins'] \
                        and 'syntax_get_ctx_content_by_id' in data['plugins']['cmd_history']:
                    # remove
                    del data['plugins']['cmd_history']['syntax_get_ctx_content_by_id']
                updated = True

            # < 2.0.149
            if old < parse_version("2.0.149"):
                print("Migrating config from < 2.0.149...")
                # logger
                if 'log.dalle' not in data:
                    data['log.dalle'] = False
                if 'log.level' not in data:
                    data['log.level'] = "error"
                if 'log.plugins' not in data:
                    data['log.plugins'] = False
                if 'log.assistants' not in data:
                    data['log.assistants'] = False
                if 'log.llama' not in data:
                    if 'llama.log' in data:
                        data['log.llama'] = data['llama.log']
                        del data['llama.log']
                    else:
                        data['log.llama'] = False

                # painter
                if 'painter.brush.color' not in data:
                    data['painter.brush.color'] = 'Black'
                if 'painter.brush.mode' not in data:
                    data['painter.brush.mode'] = 'brush'
                if 'painter.brush.size' not in data:
                    data['painter.brush.size'] = 3
                updated = True

            # < 2.0.152
            if old < parse_version("2.0.152"):
                print("Migrating config from < 2.0.152...")
                data['cmd.prompt'] = self.window.core.config.get_base('cmd.prompt')  # bg run fix
                updated = True

            # < 2.0.153
            if old < parse_version("2.0.153"):
                print("Migrating config from < 2.0.153...")
                if 'layout.dialog.geometry.store' not in data:
                    data['layout.dialog.geometry.store'] = True
                updated = True

            # < 2.0.157
            if old < parse_version("2.0.157"):
                # decrease chunk size
                print("Migrating config from < 2.0.157...")
                if 'cmd_web_google' in data['plugins'] \
                        and 'chunk_size' in data['plugins']['cmd_web_google']:
                    if data['plugins']['cmd_web_google']['chunk_size'] > 20000:
                        data['plugins']['cmd_web_google']['chunk_size'] = 20000
                updated = True

            # < 2.0.161
            if old < parse_version("2.0.161"):
                print("Migrating config from < 2.0.161...")
                if 'ctx.search_content' not in data:
                    data['ctx.search_content'] = False
                if 'download.dir' not in data:
                    data['download.dir'] = "download"
                updated = True

            # < 2.0.162 - migrate indexes into db
            if old < parse_version("2.0.162"):
                print("Migrating indexes from < 2.0.162...")
                if 'llama.idx.replace_old' not in data:
                    data['llama.idx.replace_old'] = True
                self.window.core.idx.patch(old)
                updated = True

            # < 2.0.164 - migrate indexes into db
            if old < parse_version("2.0.164"):
                print("Migrating config from < 2.0.164...")

                # Migrate plugins to provider-based versions

                # rename cmd_web_google to cmd_web
                if 'cmd_web_google' in data["plugins"]:
                    data["plugins"]["cmd_web"] = data["plugins"]["cmd_web_google"]
                    del data["plugins"]["cmd_web_google"]
                if 'cmd_web_google' in data["plugins_enabled"]:
                    data["plugins_enabled"]["cmd_web"] = data["plugins_enabled"]["cmd_web_google"]
                    del data["plugins_enabled"]["cmd_web_google"]

                # rename audio_openai_whisper to audio_input
                if 'audio_openai_whisper' in data["plugins"]:
                    data["plugins"]["audio_input"] = data["plugins"]["audio_openai_whisper"]
                    del data["plugins"]["audio_openai_whisper"]
                if 'audio_openai_whisper' in data["plugins_enabled"]:
                    data["plugins_enabled"]["audio_input"] = data["plugins_enabled"]["audio_openai_whisper"]
                    del data["plugins_enabled"]["audio_openai_whisper"]

                # migrate model to whisper_model
                if 'audio_input' in data["plugins"] and "model" in data["plugins"]["audio_input"]:
                    data["plugins"]["audio_input"]["whisper_model"] = data["plugins"]["audio_input"]["model"]
                    del data["plugins"]["audio_input"]["model"]

                # rename audio_openai_tts to audio_output
                if 'audio_openai_tts' in data["plugins"]:
                    data["plugins"]["audio_output"] = data["plugins"]["audio_openai_tts"]
                    del data["plugins"]["audio_openai_tts"]
                if 'audio_openai_tts' in data["plugins_enabled"]:
                    data["plugins_enabled"]["audio_output"] = data["plugins_enabled"]["audio_openai_tts"]
                    del data["plugins_enabled"]["audio_openai_tts"]

                # migrate model and voice to openai_model and openai_voice
                if 'audio_output' in data["plugins"] and "model" in data["plugins"]["audio_output"]:
                    data["plugins"]["audio_output"]["openai_model"] = data["plugins"]["audio_output"]["model"]
                    del data["plugins"]["audio_output"]["model"]
                if 'audio_output' in data["plugins"] and "voice" in data["plugins"]["audio_output"]:
                    data["plugins"]["audio_output"]["openai_voice"] = data["plugins"]["audio_output"]["voice"]
                    del data["plugins"]["audio_output"]["voice"]

                # migrate azure settings
                if 'audio_azure' in data["plugins"] and "azure_api_key" in data["plugins"]["audio_azure"]:
                    data["plugins"]["audio_output"]["azure_api_key"] = data["plugins"]["audio_azure"]["azure_api_key"]
                if 'audio_azure' in data["plugins"] and "azure_region" in data["plugins"]["audio_azure"]:
                    data["plugins"]["audio_output"]["azure_region"] = data["plugins"]["audio_azure"]["azure_region"]
                if 'audio_azure' in data["plugins"] and "voice_en" in data["plugins"]["audio_azure"]:
                    data["plugins"]["audio_output"]["azure_voice_en"] = data["plugins"]["audio_azure"]["voice_en"]
                if 'audio_azure' in data["plugins"] and "voice_pl" in data["plugins"]["audio_azure"]:
                    data["plugins"]["audio_output"]["azure_voice_pl"] = data["plugins"]["audio_azure"]["voice_pl"]

                # remove audio voice
                if 'audio_output' in data["plugins"] and "voice_en" in data["plugins"]["audio_output"]:
                    del data["plugins"]["audio_output"]["voice_en"]
                if 'audio_output' in data["plugins"] and "voice_pl" in data["plugins"]["audio_output"]:
                    del data["plugins"]["audio_output"]["voice_pl"]

                # remove audio_azure
                if 'audio_azure' in data["plugins"]:
                    del data["plugins"]["audio_azure"]
                if 'audio_azure' in data["plugins_enabled"]:
                    del data["plugins_enabled"]["audio_azure"]

                updated = True

            # < 2.0.165
            if old < parse_version("2.0.165"):
                print("Migrating config from < 2.0.165...")
                if 'llama.idx.excluded_ext' not in data:
                    data['llama.idx.excluded_ext'] = "3g2,3gp,7z,a,aac,aiff,alac,apk,apk,apng,app,ar,avi,avif," \
                                                     "bin,bmp,bz2,cab,class,deb,deb,dll,dmg,dmg,drv,dsd,dylib," \
                                                     "dylib,ear,egg,elf,esd,exe,flac,flv,gif,gz,heic,heif,ico," \
                                                     "img,iso,jar,jpeg,jpg,ko,lib,lz,lz4,m2v,m4a,m4v,mkv,mov,mp3," \
                                                     "mp4,mpc,msi,nrg,o,ogg,ogv,pcm,pkg,pkg,png,psd,pyc,rar,rpm,rpm," \
                                                     "so,so,svg,swm,sys,tar,tiff,vdi,vhd,vhdx,vmdk,vob,war,wav," \
                                                     "webm,webp,whl,wim,wma,wmv,xz,zip,zst"
                if 'cmd_custom' in data["plugins"]:
                    if 'cmds' in data["plugins"]["cmd_custom"]:
                        for i, cmd in enumerate(data["plugins"]["cmd_custom"]["cmds"]):
                            if "enabled" not in cmd:
                                data["plugins"]["cmd_custom"]["cmds"][i]["enabled"] = True
                updated = True

            # < 2.0.166 - migrate self_loop plugin to agent
            if old < parse_version("2.0.166"):
                print("Migrating config from < 2.0.166...")
                if 'self_loop' in data["plugins"]:
                    data["plugins"]["agent"] = data["plugins"]["self_loop"]
                    del data["plugins"]["self_loop"]
                if 'self_loop' in data["plugins_enabled"]:
                    data["plugins_enabled"]["agent"] = data["plugins_enabled"]["self_loop"]
                    del data["plugins_enabled"]["self_loop"]
                updated = True

            # < 2.0.170 - add audio input language
            if old < parse_version("2.0.170"):
                print("Migrating config from < 2.0.170...")
                if 'audio_input' in data["plugins"]:
                    # add language to google_args if not present
                    is_lang = False
                    if "google_args" in data["plugins"]["audio_input"] \
                            and isinstance(data["plugins"]["audio_input"]["google_args"], list):
                        for option in data["plugins"]["audio_input"]["google_args"]:
                            if option["name"] == "language":
                                is_lang = True
                                break
                    if not is_lang:
                        if "google_args" not in data["plugins"]["audio_input"] or \
                                not isinstance(data["plugins"]["audio_input"]["google_args"], list):
                            data["plugins"]["audio_input"]["google_args"] = []
                        data["plugins"]["audio_input"]["google_args"].append({
                            "name": "language",
                            "value": "en-US",
                            "type": "str",
                        })

                    # add language to google_cloud_args if not present
                    is_lang = False
                    if "google_cloud_args" in data["plugins"]["audio_input"] \
                            and isinstance(data["plugins"]["audio_input"]["google_cloud_args"], list):
                        for option in data["plugins"]["audio_input"]["google_cloud_args"]:
                            if option["name"] == "language":
                                is_lang = True
                                break
                    if not is_lang:
                        if "google_cloud_args" not in data["plugins"]["audio_input"] or \
                                not isinstance(data["plugins"]["audio_input"]["google_cloud_args"], list):
                            data["plugins"]["audio_input"]["google_cloud_args"] = []
                        data["plugins"]["audio_input"]["google_cloud_args"].append({
                            "name": "language",
                            "value": "en-US",
                            "type": "str",
                        })

                    # add language to bing_args if not present
                    is_lang = False
                    if "bing_args" in data["plugins"]["audio_input"] \
                            and isinstance(data["plugins"]["audio_input"]["bing_args"], list):
                        for option in data["plugins"]["audio_input"]["bing_args"]:
                            if option["name"] == "language":
                                is_lang = True
                                break
                    if not is_lang:
                        if "bing_args" not in data["plugins"]["audio_input"] or \
                                not isinstance(data["plugins"]["audio_input"]["bing_args"], list):
                            data["plugins"]["audio_input"]["bing_args"] = []
                        data["plugins"]["audio_input"]["bing_args"].append({
                            "name": "language",
                            "value": "en-US",
                            "type": "str",
                        })
                updated = True

            # < 2.0.172 - fix cmd syntax
            if old < parse_version("2.0.172"):
                print("Migrating config from < 2.0.172...")
                if 'cmd_files' in data["plugins"] and 'syntax_file_index' in data["plugins"]["cmd_files"]:
                    syntax = '"file_index": use it to index (embed in Vector Store) a file or directory for ' \
                             'future use, params: "path"'
                    data["plugins"]["cmd_files"]["syntax_file_index"] = syntax
                if 'cmd_web' in data["plugins"] and 'syntax_web_url_open' in data["plugins"]["cmd_web"]:
                    syntax = '"web_url_open": use it to get contents from a specific Web page. ' \
                             'Use a custom summary prompt if necessary, otherwise a default summary will be used, ' \
                             'params: "url", "summarize_prompt"'
                    data["plugins"]["cmd_web"]["syntax_web_url_open"] = syntax
                if 'cmd_web' in data["plugins"] and 'syntax_web_url_raw' in data["plugins"]["cmd_web"]:
                    syntax = '"web_url_raw": use it to get RAW text/html content (not summarized) from ' \
                             'a specific Web page, params: "url"'
                    data["plugins"]["cmd_web"]["syntax_web_url_raw"] = syntax
                if 'cmd_web' in data["plugins"] and 'syntax_web_urls' in data["plugins"]["cmd_web"]:
                    syntax = '"web_urls": use it to search the Web for URLs to use, prepare a search query itself, ' \
                             'a list of found links to websites will be returned, 10 links per page max. ' \
                             'You can change the page or the number of links per page using the provided parameters, ' \
                             'params: "query", "page", "num_links"'
                    data["plugins"]["cmd_web"]["syntax_web_urls"] = syntax
                updated = True

            # < 2.1.1
            if old < parse_version("2.1.1"):
                print("Migrating config from < 2.1.1...")
                if 'llama.hub.loaders.args' not in data:
                    data['llama.hub.loaders.args'] = []
                updated = True

            # < 2.1.2
            if old < parse_version("2.1.2"):
                print("Migrating config from < 2.1.2...")
                if 'llama.hub.loaders.use_local' not in data:
                    data['llama.hub.loaders.use_local'] = False
                updated = True

            # < 2.1.5 - syntax
            if old < parse_version("2.1.5"):
                print("Migrating config from < 2.1.5...")
                if 'cmd_files' in data["plugins"] and 'syntax_file_index' in data["plugins"]["cmd_files"]:
                    syntax = '"file_index": use it to index (embed in Vector Store) a file or directory, params: "path"'
                    data["plugins"]["cmd_files"]["syntax_file_index"] = syntax
                    updated = True

            # < 2.1.8 - syntax
            if old < parse_version("2.1.8"):
                print("Migrating config from < 2.1.8...")
                if 'idx_llama_index' in data["plugins"] and 'syntax_prepare_question' in data["plugins"]["idx_llama_index"]:
                    syntax = 'Simplify the question into a short query for retrieving information from a vector store.'
                    data["plugins"]["idx_llama_index"]["syntax_prepare_question"] = syntax
                    updated = True

            # < 2.1.9 - syntax
            if old < parse_version("2.1.9"):
                print("Migrating config from < 2.1.9...")
                if 'cmd_files' in data["plugins"] and 'syntax_read_file' in data["plugins"]["cmd_files"]:
                    syntax = '"read_file": read data from file, if multiple files then pass list of files, params: "filename"'
                    data["plugins"]["cmd_files"]["syntax_read_file"] = syntax
                if 'log.events' not in data:
                    data["log.events"] = False
                    updated = True

            # < 2.1.10
            if old < parse_version("2.1.10"):
                print("Migrating config from < 2.1.10...")

                # fix missing updated before >>>
                if 'cmd_files' in data["plugins"] and 'syntax_file_index' in data["plugins"]["cmd_files"]:
                    syntax = '"file_index": use it to index (embed in Vector Store) a file or directory, params: "path"'
                    data["plugins"]["cmd_files"]["syntax_file_index"] = syntax
                if 'idx_llama_index' in data["plugins"] and 'syntax_prepare_question' in data["plugins"]["idx_llama_index"]:
                    syntax = 'Simplify the question into a short query for retrieving information from a vector store.'
                    data["plugins"]["idx_llama_index"]["syntax_prepare_question"] = syntax
                if 'cmd_files' in data["plugins"] and 'syntax_read_file' in data["plugins"]["cmd_files"]:
                    syntax = '"read_file": read data from file, if multiple files then pass list of files, params: "filename"'
                    data["plugins"]["cmd_files"]["syntax_read_file"] = syntax
                if 'log.events' not in data:
                    data["log.events"] = False
                # <<< fix missing updated before

                # current
                if 'ctx.records.filter.labels' not in data:
                    data["ctx.records.filter.labels"] = [0, 1, 2, 3, 4, 5, 6, 7]

                if 'preset.plugins' not in data:
                    data["preset.plugins"] = ""

                if 'ctx.records.filter' in data and str(data["ctx.records.filter"]).startswith("label"):
                    data["ctx.records.filter"] = "all"
                updated = True

            if old < parse_version("2.1.12"):
                print("Migrating config from < 2.1.12...")
                if 'max_requests_limit' not in data:
                    data["max_requests_limit"] = 60
                if 'ctx.allow_item_delete' not in data:
                    data["ctx.allow_item_delete"] = True
                if 'ctx.counters.all' not in data:
                    data["ctx.counters.all"] = False
                if 'agent.prompt.continue' not in data:
                    data["agent.prompt.continue"] = "continue..."
                if 'api_endpoint' not in data:
                    data["api_endpoint"] = "https://api.openai.com/v1"
                updated = True

            if old < parse_version("2.1.15"):
                print("Migrating config from < 2.1.15...")
                if 'ctx.edit_icons' not in data:
                    data["ctx.edit_icons"] = False
                if 'llama.idx.auto.modes' not in data:
                    data["llama.idx.auto.modes"] = "chat,completion,vision,assistant,langchain,llama_index,agent"
                if 'ctx.allow_item_delete' in data:
                    del data["ctx.allow_item_delete"]
                updated = True

            if old < parse_version("2.1.16"):
                print("Migrating config from < 2.1.16...")
                if 'ctx.sources' not in data:
                    data["ctx.sources"] = True
                updated = True

            if old < parse_version("2.1.18"):
                print("Migrating config from < 2.1.18...")
                if 'ctx.audio' not in data:
                    data["ctx.audio"] = True
                updated = True

            if old < parse_version("2.1.19"):
                print("Migrating config from < 2.1.19...")
                if 'llama.idx.excluded_ext' in data:
                    data["llama.idx.excluded.ext"] = copy.deepcopy(data["llama.idx.excluded_ext"])
                    del data["llama.idx.excluded_ext"]
                if 'llama.idx.excluded.force' not in data:
                    data["llama.idx.excluded.force"] = False
                if 'llama.idx.custom_meta' not in data:
                    data["llama.idx.custom_meta"] = [
                          {
                              "extensions": "*",
                              "key": "file_name",
                              "value": "{relative_path}"
                          }
                    ]
                updated = True

            # < 2.1.20
            if old < parse_version("2.1.20"):
                print("Migrating config from < 2.1.20...")
                data['cmd.prompt'] = self.window.core.config.get_base('cmd.prompt')  # moved to json schema
                updated = True

            # < 2.1.22
            if old < parse_version("2.1.22"):
                print("Migrating config from < 2.1.22...")
                if 'llama.idx.custom_meta.web' not in data:
                    data["llama.idx.custom_meta.web"] = []
                updated = True

            # < 2.1.23
            if old < parse_version("2.1.23"):
                print("Migrating config from < 2.1.23...")
                if 'llama.idx.embeddings.provider' not in data:
                    data["llama.idx.embeddings.provider"] = "openai"
                if 'llama.idx.embeddings.args' not in data:
                    data["llama.idx.embeddings.args"] = [
                        {
                            "name": "model",
                            "value": "text-embedding-3-small",
                            "type": "str"
                        }
                    ]
                if 'llama.idx.embeddings.env' not in data:
                    data["llama.idx.embeddings.env"] = [
                        {
                            "name": "OPENAI_API_KEY",
                            "value": "{api_key}",
                        },
                        {
                            "name": "OPENAI_API_BASE",
                            "value": "{api_endpoint}",
                        }
                    ]
                self.window.core.plugins.reset_options("cmd_web", [
                    "cmd.web_url_open",
                    "cmd.web_url_raw",
                ])
                updated = True

            # < 2.1.26
            if old < parse_version("2.1.26"):
                print("Migrating config from < 2.1.26...")
                if 'agent.prompt.continue' in data and data['agent.prompt.continue'] == 'continue...':
                    data["agent.prompt.continue"] = "continue if needed..."
                self.window.core.plugins.reset_options("idx_llama_index", [
                    "prompt",
                ])
                updated = True

            # < 2.1.28
            if old < parse_version("2.1.28"):
                print("Migrating config from < 2.1.28...")
                if 'log.ctx' not in data:
                    data["log.ctx"] = True
                if 'llama.idx.embeddings.limit.rpm' not in data:
                    data["llama.idx.embeddings.limit.rpm"] = 100
                updated = True

            # < 2.1.29
            if old < parse_version("2.1.29"):
                print("Migrating config from < 2.1.29...")
                self.window.core.plugins.reset_options("cmd_code_interpreter", [
                    "cmd.code_execute",
                ])
                updated = True

            # < 2.1.31
            if old < parse_version("2.1.31"):
                print("Migrating config from < 2.1.31...")
                self.window.core.plugins.reset_options("cmd_code_interpreter", [
                    "cmd.code_execute",
                ])
                files_to_move = [
                    {"_interpreter.current.py": ".interpreter.current.py"},
                    {"_interpreter.py": ".interpreter.output.py"},
                    {"_interpreter.input.py": ".interpreter.input.py"},
                ]
                files_to_remove = [
                    "_interpreter.tmp.py",
                ]
                dir = self.window.core.config.get_user_dir("data")
                try:
                    for file in files_to_move:
                        for src, dst in file.items():
                            src = os.path.join(dir, src)
                            dst = os.path.join(dir, dst)
                            if os.path.exists(src):
                                os.rename(src, dst)
                    for file in files_to_remove:
                        file = os.path.join(dir, file)
                        if os.path.exists(file):
                            os.remove(file)
                except Exception as e:
                    print("Error while migrating interpreter files:", e)

                updated = True

            # < 2.1.32
            if old < parse_version("2.1.32"):
                print("Migrating config from < 2.1.32...")
                if 'ctx.use_extra' not in data:
                    data["ctx.use_extra"] = True

                self.window.core.plugins.reset_options("real_time", [
                    "tpl",
                ])

                # old keys first
                data['cmd.prompt'] = self.window.core.config.get_base('prompt.cmd')  # new format
                data['cmd.prompt.extra'] = self.window.core.config.get_base('prompt.cmd.extra')  # new format
                data['cmd.prompt.extra.assistants'] = self.window.core.config.get_base('prompt.cmd.extra.assistants')  # new format

                # new keys
                data['prompt.agent.goal'] = self.window.core.config.get_base('prompt.agent.goal')  # new format

                # replace to new keys
                self.window.core.config.replace_key(data, "img_prompt", "prompt.img")
                self.window.core.config.replace_key(data, "ctx.auto_summary.prompt", "prompt.ctx.auto_summary.user")
                self.window.core.config.replace_key(data, "ctx.auto_summary.system", "prompt.ctx.auto_summary.system")
                self.window.core.config.replace_key(data, "cmd.prompt", "prompt.cmd")
                self.window.core.config.replace_key(data, "cmd.prompt.extra", "prompt.cmd.extra")
                self.window.core.config.replace_key(data, "cmd.prompt.extra.assistants", "prompt.cmd.extra.assistants")
                self.window.core.config.replace_key(data, "agent.prompt.continue", "prompt.agent.continue")
                self.window.core.config.replace_key(data, "default_prompt", "prompt.default")
                updated = True

            # < 2.1.35
            if old < parse_version("2.1.35"):
                print("Migrating config from < 2.1.35...")
                if 'interpreter.auto_clear' not in data:
                    data["interpreter.auto_clear"] = True
                if 'interpreter.execute_all' not in data:
                    data["interpreter.execute_all"] = True
                if 'interpreter.edit' not in data:
                    data["interpreter.edit"] = False
                if 'interpreter.input' not in data:
                    data["interpreter.input"] = ""
                if 'video.player.path' not in data:
                    data["video.player.path"] = ""
                if 'video.player.volume' not in data:
                    data["video.player.volume"] = 100
                if 'video.player.volume.mute' not in data:
                    data["video.player.volume.mute"] = False
                updated = True

            # < 2.1.37
            if old < parse_version("2.1.37"):
                print("Migrating config from < 2.1.37...")
                if 'audio.transcribe.convert_video' not in data:
                    data["audio.transcribe.convert_video"] = True
                updated = True

            # < 2.1.38
            if old < parse_version("2.1.38"):
                print("Migrating config from < 2.1.38...")
                if 'lang' in data and data['lang'] == 'ua':
                    data["lang"] = "uk"
                updated = True

            # < 2.1.45
            if old < parse_version("2.1.45"):
                print("Migrating config from < 2.1.45...")
                if 'ctx.copy_code' not in data:
                    data["ctx.copy_code"] = True
                updated = True

            # < 2.1.52
            if old < parse_version("2.1.52"):
                print("Migrating config from < 2.1.52...")
                if 'ctx.code_interpreter' not in data:
                    data["ctx.code_interpreter"] = True
                updated = True

            # < 2.1.59
            if old < parse_version("2.1.59"):
                print("Migrating config from < 2.1.59...")
                if 'render.code_syntax' not in data:
                    data["render.code_syntax"] = "github-dark"
                if 'zoom' not in data:
                    data["zoom"] = 1.0
                if 'ctx.convert_lists' not in data:
                    data["ctx.convert_lists"] = False
                if 'render.engine' not in data:
                    data["render.engine"] = "web"
                if 'render.open_gl' not in data:
                    data["render.open_gl"] = False

                # in snap, leave legacy render engine by default
                # if self.window.core.platforms.is_snap():
                    # data["render.engine"] = "legacy"

                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # NEW
                self.window.core.updater.patch_css('web.light.css', True)  # NEW
                self.window.core.updater.patch_css('web.dark.css', True)  #  NEW
                updated = True

            # < 2.1.60
            if old < parse_version("2.1.60"):
                print("Migrating config from < 2.1.60...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                updated = True

            # < 2.1.61
            if old < parse_version("2.1.61"):
                print("Migrating config from < 2.1.61...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                updated = True

            # < 2.1.63
            if old < parse_version("2.1.63"):
                print("Migrating config from < 2.1.63...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.70
            if old < parse_version("2.1.70"):
                print("Migrating config from < 2.1.70...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                updated = True

            # < 2.1.72
            if old < parse_version("2.1.72"):
                print("Migrating config from < 2.1.72...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.73
            if old < parse_version("2.1.73"):
                print("Migrating config from < 2.1.73...")
                # fix: issue #50
                if "llama.idx.embeddings.args" in data:
                    for arg in data["llama.idx.embeddings.args"]:
                        if "type" not in arg:
                            arg["type"] = "str"
                updated = True

            # < 2.1.74
            if old < parse_version("2.1.74"):
                print("Migrating config from < 2.1.74...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.75
            if old < parse_version("2.1.75"):
                print("Migrating config from < 2.1.75...")
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.76
            if old < parse_version("2.1.76"):
                print("Migrating config from < 2.1.76...")
                if 'render.blocks' not in data:
                    data["render.blocks"] = True
                # css upgrade
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.78
            if old < parse_version("2.1.78"):
                print("Migrating config from < 2.1.78...")
                # css upgrade, scroll bg
                self.window.core.updater.patch_css('web.css', True)  # force update
                self.window.core.updater.patch_css('web.light.css', True)  # force update
                self.window.core.updater.patch_css('web.dark.css', True)  # force update
                updated = True

            # < 2.1.79
            if old < parse_version("2.1.79"):
                print("Migrating config from < 2.1.79...")
                if 'assistant.store.hide_threads' not in data:
                    data["assistant.store.hide_threads"] = True
                updated = True

            # < 2.2.2
            if old < parse_version("2.2.2"):
                print("Migrating config from < 2.2.2...")
                if 'app.env' not in data:
                    data["app.env"] = []
                updated = True

            # < 2.2.7
            if old < parse_version("2.2.7"):
                print("Migrating config from < 2.2.7...")
                if 'prompt.agent.instruction' not in data:
                    data["prompt.agent.instruction"] = ""
                if 'prompt.expert' not in data:
                    data["prompt.expert"] = ""
                if 'experts.mode' not in data:
                    data["experts.mode"] = "chat"
                # from base
                data["prompt.agent.instruction"] = self.window.core.config.get_base('prompt.agent.instruction')
                data["prompt.agent.continue"] = self.window.core.config.get_base('prompt.agent.continue')
                data["prompt.agent.goal"] = self.window.core.config.get_base('prompt.agent.goal')
                data["prompt.expert"] = self.window.core.config.get_base('prompt.expert')
                data["prompt.img"] = self.window.core.config.get_base('prompt.img')
                updated = True

            # < 2.2.8
            if old < parse_version("2.2.8"):
                print("Migrating config from < 2.2.8...")
                if 'access.audio.event.speech' not in data:
                    data["access.audio.event.speech"] = False
                if 'access.audio.notify.execute' not in data:
                    data["access.audio.notify.execute"] = True
                if 'access.microphone.notify' not in data:
                    data["access.microphone.notify"] = True
                if 'access.shortcuts' not in data:
                    data["access.shortcuts"] = [
                        {
                            "action": "voice_cmd.toggle",
                            "key": "Space",
                            "key_modifier": "Control"
                        },
                        {
                            "action": "tab.chat",
                            "key": "1",
                            "key_modifier": "Control"
                        },
                        {
                            "action": "tab.files",
                            "key": "2",
                            "key_modifier": "Control"
                        },
                        {
                            "action": "tab.calendar",
                            "key": "3",
                            "key_modifier": "Control"
                        },
                        {
                            "action": "tab.draw",
                            "key": "4",
                            "key_modifier": "Control"
                        },
                        {
                            "action": "tab.notepad",
                            "key": "5",
                            "key_modifier": "Control"
                        }
                    ]
                if 'access.voice_control' not in data:
                    data["access.voice_control"] = False
                if 'access.voice_control.model' not in data:
                    data["access.voice_control.model"] = "gpt-3.5-turbo"
                updated = True

        # update file
        migrated = False
        if updated:
            data = dict(sorted(data.items()))
            self.window.core.config.data = data
            self.window.core.config.save()
            migrated = True

        # check for any missing config keys if versions mismatch
        if is_old:
            if self.window.core.updater.post_check_config():
                migrated = True

        return migrated



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.26 15:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QVBoxLayout, QMenuBar

from pygpt_net.tools.media_player.ui.widgets import VideoPlayerWidget
from pygpt_net.ui.widget.dialog.base import BaseDialog

from pygpt_net.utils import trans

class VideoPlayer:
    def __init__(self, window=None):
        """
        Video Player dialog

        :param window: Window instance
        """
        self.window = window
        self.path = None
        self.menu_bar = None
        self.file_menu = None
        self.actions = {}

    def setup_menu(self) -> QMenuBar:
        """Setup video dialog menu"""
        self.menu_bar = QMenuBar()
        self.file_menu = self.menu_bar.addMenu(trans("menu.file"))

        # open
        self.actions["open"] = QAction(QIcon(":/icons/folder.svg"), trans("action.open"))
        self.actions["open"].triggered.connect(
            lambda: self.window.tools.get("player").open_file()
        )

        # save as
        self.actions["save_as"] = QAction(QIcon(":/icons/save"), trans("action.save_as"))
        self.actions["save_as"].triggered.connect(
            lambda: self.window.tools.get("player").save_as_file()
        )

        # exit
        self.actions["exit"] = QAction(QIcon(":/icons/logout.svg"), trans("menu.file.exit"))
        self.actions["exit"].triggered.connect(
            lambda: self.window.tools.get("player").close()
        )

        # add actions
        self.file_menu.addAction(self.actions["open"])
        self.file_menu.addAction(self.actions["save_as"])
        self.file_menu.addAction(self.actions["exit"])
        return self.menu_bar

    def setup(self):
        """Setup video dialog"""
        id = 'video_player'
        self.window.video_player = VideoPlayerWidget(self.window)

        layout = QVBoxLayout()
        layout.setMenuBar(self.setup_menu())
        layout.addWidget(self.window.video_player)

        self.window.ui.dialog[id] = VideoPlayerDialog(self.window, id)
        self.window.ui.dialog[id].setLayout(layout)
        self.window.ui.dialog[id].setWindowTitle("Media Player")

class VideoPlayerDialog(BaseDialog):
    def __init__(self, window=None, id=None):
        """
        VideoPlayer dialog

        :param window: main window
        :param id: info window id
        """
        super(VideoPlayerDialog, self).__init__(window, id)
        self.window = window
        self.id = id

    def closeEvent(self, event):
        """
        Close event

        :param event: close event
        """
        self.cleanup()
        super(VideoPlayerDialog, self).closeEvent(event)

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key press event
        """
        if event.key() == Qt.Key_Escape:
            self.cleanup()
            self.close()  # close dialog when the Esc key is pressed.
        else:
            super(VideoPlayerDialog, self).keyPressEvent(event)

    def cleanup(self):
        """
        Cleanup on close
        """
        self.window.tools.get("player").on_close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 04:00:00                  #
# ================================================== #

from openai import AssistantEventHandler

from PySide6.QtCore import QObject, Signal, QRunnable, Slot
from typing_extensions import override

from pygpt_net.item.ctx import CtxItem


class AssistantsWorker:
    CHUNK_TYPE_MSG = 1  # last chunk = message
    CHUNK_TYPE_TOOL = 0  # last chunk = tool

    def __init__(self, window=None):
        """
        Assistants Worker (async)

        :param window: Window instance
        """
        self.window = window
        self.tool_output = ""
        self.last_chunk_type = self.CHUNK_TYPE_MSG  # msg or tool

    def create_run(
            self,
            ctx: CtxItem,
            thread_id: str,
            assistant_id: str,
            system_prompt: str
    ):
        """
        Create assistant run

        :param ctx: context item
        :param thread_id: thread id
        :param assistant_id: assistant id
        :param system_prompt: system prompt
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "run_create"
        worker.ctx = ctx
        worker.thread_id = thread_id
        worker.assistant_id = assistant_id
        worker.system_prompt = system_prompt
        worker.stream = self.window.core.config.get("stream")
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def msg_send(
            self,
            ctx: CtxItem,
            thread_id: str,
            assistant_id: str,
            model: str,
            file_ids: list,
            prompt: str,
            system_prompt: str
    ):
        """
        Send message to assistant thread

        :param ctx: context item
        :param thread_id: thread id
        :param assistant_id: assistant id
        :param model: model ID
        :param file_ids: uploaded file IDs
        :param prompt: prompt
        :param system_prompt: system prompt
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "msg_send"
        worker.model = model
        worker.ctx = ctx
        worker.thread_id = thread_id
        worker.assistant_id = assistant_id
        worker.prompt = prompt
        worker.file_ids = file_ids
        worker.system_prompt = system_prompt
        worker.stream = self.window.core.config.get("stream")
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def tools_submit(
            self,
            ctx: CtxItem,
            model: str,
            tools_outputs: list,
    ):
        """
        Send tools outputs to assistant thread

        :param ctx: context item
        :param model: model ID
        :param tools_outputs: list of tools outputs
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "tools_submit"
        worker.model = model
        worker.ctx = ctx
        worker.tools_outputs = tools_outputs
        worker.stream = self.window.core.config.get("stream")
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def connect_signals(self, worker):
        """
        Connect worker signals

        :param worker: worker instance
        """
        worker.signals.finished.connect(self.handle_run_created)
        worker.signals.error.connect(self.handle_error)
        worker.signals.stream_text_created.connect(self.handle_text_created)
        worker.signals.stream_text_delta.connect(self.handle_text_delta)
        worker.signals.stream_end.connect(self.handle_end)
        worker.signals.stream_tool_call_created.connect(self.handle_tool_call_created)
        worker.signals.stream_tool_call_delta.connect(self.handle_tool_call_delta)
        worker.signals.stream_tool_call_done.connect(self.handle_tool_call_done)
        worker.signals.stream_message_done.connect(self.handle_message_done)
        worker.signals.stream_run_step_created.connect(self.handle_run_step_created)
        worker.signals.stream_run_step_done.connect(self.handle_run_step_done)
        worker.signals.stream_run_created.connect(self.handle_run_created)

    def is_show_tool_output(self) -> bool:
        """
        Return True if show tool output

        :return: bool
        """
        return self.window.core.config.get('ctx.code_interpreter', False)

    def begin_chunks(self, ctx: CtxItem):
        """
        Begin chunks stream output

        :param ctx: context item
        """
        self.window.controller.chat.render.stream_begin()
        self.window.controller.assistant.threads.handle_stream_begin(ctx)
        self.window.controller.chat.render.append_chunk(
            ctx,
            "",
            True,
        )  # only begin chunks

    def reset_tool_output(self):
        """Reset tool output"""
        self.tool_output = ""

    def has_tool_output(self) -> bool:
        """
        Return True if has tool output

        :return: bool
        """
        return self.tool_output is not None and self.tool_output != ""

    def append_chunk(self, ctx: CtxItem, chunk: str, begin: bool = False):
        """
        Append chunk to context output

        :param ctx: context item
        :param chunk: chunk
        :param begin: stream text begin
        """
        if ctx.output is None:
            ctx.output = ""
        if chunk is not None:
            ctx.output += chunk
        self.window.controller.chat.render.append_chunk(
            ctx,
            chunk,
            False,
        )

    def log(self, message: str):
        """
        Log assistant message

        :param message: message
        """
        self.window.controller.assistant.threads.log(message)

    @Slot(object)
    def handle_text_created(self, ctx: CtxItem):
        """
        Handle thread text created signal

        :param ctx: context item
        """
        self.log("STREAM: text created")
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return
        if (ctx.output != "" and ctx.output is not None) or self.has_tool_output():
            return  # append to existing output

    @Slot(object, object)
    def handle_text_delta(self, ctx: CtxItem, delta, begin):
        """
        Handle thread text delta signal

        :param ctx: context item
        :param delta: delta
        :param begin: stream text begin
        """
        if ctx.stopped:
            return
        # self.window.controller.assistant.threads.log("STREAM: text delta")
        if ctx.output_tokens is None:
            ctx.output_tokens = 0
        ctx.output_tokens += 1  # tokens++
        self.append_chunk(ctx, delta.value, begin)
        self.last_chunk_type = self.CHUNK_TYPE_MSG
        # self.log("STREAM: chunk: msg text")

    @Slot(object)
    def handle_end(self, ctx: CtxItem):
        """
        Handle thread end signal

        :param ctx: context item
        """
        self.log("STREAM: end")
        if ctx.stopped:
            return
        self.reset_tool_output()
        self.window.controller.chat.render.stream_end()
        self.window.controller.assistant.threads.handle_output_message(ctx, stream=True)

    @Slot(object, object, bool)
    def handle_tool_call_created(self, ctx: CtxItem, tool_call, begin):
        """
        Handle thread tool call created signal

        :param ctx: context item
        :param tool_call: tool call
        :param begin: stream text begin
        """
        self.log("STREAM: tool call created: {}".format(tool_call.type))
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return
        if self.is_show_tool_output():
            self.reset_tool_output()

    @Slot(object, object, bool)
    def handle_tool_call_delta(self, ctx: CtxItem, delta, begin):
        """
        Handle thread tool call delta signal

        :param ctx: context item
        :param delta: delta
        :param begin: stream text begin
        """
        if ctx.stopped:
            return
        # self.log("STREAM: tool call delta")
        if self.is_show_tool_output():
            if delta.type == 'code_interpreter':
                # begin code block
                if begin:
                    self.log("STREAM: tool output: code_interpreter: begin")
                    self.append_chunk(ctx, "\n\n**Code interpreter**\n```python\n", begin=False)

                # input
                if delta.code_interpreter.input:
                    if delta.code_interpreter.input is not None:
                        self.tool_output += delta.code_interpreter.input
                        self.append_chunk(ctx, delta.code_interpreter.input)
                        self.last_chunk_type = self.CHUNK_TYPE_TOOL
                        # self.log("STREAM: chunk: code_interpreter: input")

                # output
                if delta.code_interpreter.outputs:
                    header = False
                    for output in delta.code_interpreter.outputs:
                        if output.type == "logs":
                            if not header:
                                self.log("STREAM: tool output: code_interpreter: log begin")
                                self.append_chunk(ctx, "\n\n-------\nOutput:\n")
                                self.last_chunk_type = self.CHUNK_TYPE_TOOL
                                header = True
                            chunk = output.logs
                            if output.logs is not None:
                                self.append_chunk(ctx, "\n" + chunk)
                                self.last_chunk_type = self.CHUNK_TYPE_TOOL
                                # self.log("STREAM: chunk: code_interpreter: logs")

    @Slot(object, object, bool)
    def handle_tool_call_done(self, ctx: CtxItem, tool_call, begin):
        """
        Handle thread tool call done signal

        :param ctx: context item
        :param tool_call: tool call
        :param begin: stream text begin
        """
        self.log("STREAM: tool call done: {}".format(tool_call.type))
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return
        if self.is_show_tool_output():
            if self.has_tool_output() and self.last_chunk_type == self.CHUNK_TYPE_TOOL:
                self.append_chunk(ctx, "\n```\n")

    @Slot(object, object)
    def handle_message_done(self, ctx: CtxItem, message):
        """
        Handle thread message done signal

        :param ctx: context item
        :param message: message
        """
        self.log("STREAM: message done")
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return
        self.window.controller.assistant.threads.handle_message_data(
            ctx,
            message,
            stream=True,
        )  # handle img, files, etc.

    @Slot(object, object, int)
    def handle_run_step_created(self, ctx: CtxItem, run_step, step_idx: int):
        """
        Handle thread run step created signal

        :param ctx: context item
        :param run_step: run step
        :param step_idx: step index
        """
        self.log("STREAM: run step created: {}, idx: {}".format(run_step.id, step_idx))
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return
        ctx.run_id = run_step.run_id
        self.window.controller.assistant.threads.handle_run_step_created(ctx, stream=True)
        if step_idx == 0:
            self.log("STREAM: run created: {}".format(ctx.run_id))
            self.log("STREAM: begin chunks...")
            self.begin_chunks(ctx)

    @Slot(object, object, int)
    def handle_run_step_done(self, ctx: CtxItem, run_step, step_idx: int):
        """
        Handle thread run step created signal

        :param ctx: context item
        :param run_step: run step
        :param step_idx: step index
        """
        self.log("STREAM: run step done: {}, idx: {}".format(run_step.id, step_idx))
        if ctx.stopped:
            self.log("STREAM: ignoring: stopped")
            return

    @Slot(object, object)
    def handle_error(self, ctx: CtxItem, err: any):
        """
        Handle thread error signal

        :param ctx: context item
        :param err: error message
        """
        self.log("STREAM: error: {}".format(err))
        self.window.controller.assistant.threads.handle_run_error(ctx, err)

    @Slot(object, object, bool)
    def handle_run_created(self, ctx: CtxItem, run, stream=False):
        """
        Handle thread finished signal

        :param ctx: context item
        :param run
        :param stream: stream mode
        """
        if not stream:
            self.log("STREAM: run created: {}".format(run.id))
        self.window.controller.assistant.threads.handle_run_created(
            ctx,
            run,
            stream=stream
        )


class EventHandler(AssistantEventHandler):
    """Assistants event handler"""
    def __init__(
            self,
            signals: QObject = None,
            ctx: CtxItem = None

    ):
        super().__init__()
        self.signals = signals
        self.ctx = ctx
        self.text_begin = False
        self.tool_begin = False
        self.stream_started = False
        self.step_idx = 0
        
    @override
    def on_text_created(self, text) -> None:
        """Callback that is fired when a text content block is created"""
        # print(f"\nassistant > ", end="", flush=True)
        if not self.stream_started:
            self.text_begin = True
        self.stream_started = True
        self.signals.stream_text_created.emit(self.ctx)

    @override
    def on_text_delta(self, delta, snapshot):
        """Callback that is fired whenever a message delta is returned from the API"""
        # print(delta.value, end="", flush=True)
        self.signals.stream_text_delta.emit(self.ctx, delta, self.text_begin)
        self.text_begin = False

    @override
    def on_run_step_created(self, run_step) -> None:
        """Callback that is fired when a run step is created"""
        self.tool_begin = False  # reset
        self.signals.stream_run_step_created.emit(self.ctx, run_step, self.step_idx)
        self.step_idx += 1

    @override
    def on_run_step_done(self, run_step):
        """Fires when run completed."""
        self.signals.stream_run_step_done.emit(self.ctx, run_step, self.step_idx)

    @override
    def on_text_done(self, text):
        """Callback that is fired when a text content block is finished"""
        pass

    @override
    def on_message_done(self, message) -> None:
        """Callback that is fired when a message is completed"""
        self.signals.stream_message_done.emit(self.ctx, message)

    @override
    def on_end(self):
        """Fires when the stream has finished."""
        # print("\n\nassistant > end\n", flush=True)
        self.signals.stream_end.emit(self.ctx)

    @override
    def on_exception(self, exception: Exception):
        """Fired whenever an exception happens during streaming"""
        self.signals.error.emit(self.ctx,  exception)

    @override
    def on_timeout(self) -> None:
        """Fires if the request times out"""
        self.signals.error.emit(self.ctx,  "Timeout")

    @override
    def on_tool_call_created(self, tool_call):
        """Callback that is fired when a tool call is created"""
        # print(f"\nassistant > {tool_call.type}\n", flush=True)
        self.tool_begin = False  # reset
        self.signals.stream_tool_call_created.emit(self.ctx, tool_call, self.tool_begin)

    @override
    def on_tool_call_delta(self, delta, snapshot):
        """Callback that is fired when a tool call delta is encountered"""
        begin = True
        if self.tool_begin:
            begin = False
        self.signals.stream_tool_call_delta.emit(self.ctx, delta, begin)
        self.tool_begin = True

    @override
    def on_tool_call_done(self, tool_call) -> None:
        """Callback that is fired when a tool call delta is encountered"""
        self.signals.stream_tool_call_done.emit(self.ctx, tool_call, self.tool_begin)


class WorkerSignals(QObject):
    """Assistants worker signals"""
    # run create
    finished = Signal(object, object, bool)  # ctx, run, stream
    error = Signal(object, object)  # ctx, error

    # stream
    stream_text_created = Signal(object)  # ctx
    stream_text_delta = Signal(object, object, bool)  # ctx, delta, begin
    stream_end = Signal(object)  # ctx
    stream_tool_call_created = Signal(object, object, bool)  # ctx, tool_call, begin
    stream_tool_call_delta = Signal(object, object, bool)  # ctx, delta, begin
    stream_tool_call_done = Signal(object, object, bool)  # ctx, tool_call
    stream_message_done = Signal(object, object)  # ctx, message
    stream_run_step_created = Signal(object, object, int)  # ctx, run_step, step idx
    stream_run_created = Signal(object, object)  # ctx, run_step
    stream_run_step_done = Signal(object, object, int)  # ctx, run_step, step idx


class Worker(QRunnable):
    """Assistants worker"""
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.window = None
        self.mode = None
        self.model = None
        self.thread_id = None
        self.assistant_id = None
        self.file_ids = []
        self.prompt = None
        self.system_prompt = None
        self.tools_outputs = None
        self.ctx = None
        self.stream = False

    @Slot()
    def run(self):
        """Assistants worker thread"""
        if self.mode == "run_create":
            self.run_create()
        elif self.mode == "msg_send":
            self.msg_send()
        elif self.mode == "tools_submit":
            self.tools_submit()

    def run_create(self) -> bool:
        """
        Create assistant run

        :return: result
        """
        try:
            if self.stream:  # stream mode
                run = self.window.core.gpt.assistants.run_create_stream(
                    self.signals,
                    self.ctx,
                    self.thread_id,
                    self.assistant_id,
                    self.model,
                    self.system_prompt,
                )
            else:
                # not stream mode
                run = self.window.core.gpt.assistants.run_create(
                    self.thread_id,
                    self.assistant_id,
                    self.model,
                    self.system_prompt,
                )
            # handle run (stream or not)
            if run is not None:
                self.signals.finished.emit(self.ctx, run, self.stream)
                return True
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False

    def msg_send(self) -> bool:
        """
        Send message to assistant

        :return: result
        """
        try:
            response = self.window.core.gpt.assistants.msg_send(
                self.thread_id,
                self.prompt,
                self.file_ids,
            )
            if response is not None:
                self.ctx.msg_id = response.id
                return self.run_create()
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False

    def tools_submit(self) -> bool:
        """
        Submit tools outputs to assistant

        :return: result
        """
        try:
            run = self.window.core.gpt.assistants.run_submit_tool(self.ctx, self.tools_outputs)
            if run is not None:
                self.ctx.run_id = run.id  # update run id
                self.signals.finished.emit(self.ctx, run, False)  # continue status check
                # TODO: implement stream mode in tool submit
                return True
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QWidget, QHBoxLayout, QPushButton, QVBoxLayout, QLabel

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc

class ExpertsEditor(QWidget):
    def __init__(self, window=None):
        """
        Experts select menu

        :param window: main window
        """
        super(ExpertsEditor, self).__init__(window)
        self.window = window
        self.id = "preset.experts"
        self.layout = self.setup()
        self.window.ui.models["preset.experts.available"] = self.create_model(self.window)
        self.window.ui.nodes["preset.experts.available"].setModel(self.window.ui.models["preset.experts.available"])
        self.window.ui.nodes["preset.experts.available"].selectionModel().selectionChanged.connect(
            lambda: self.window.controller.presets.editor.experts.change_available()
        )
        self.window.ui.models["preset.experts.selected"] = self.create_model(self.window)
        self.window.ui.nodes["preset.experts.selected"].setModel(self.window.ui.models["preset.experts.selected"])
        self.window.ui.nodes["preset.experts.selected"].selectionModel().selectionChanged.connect(
            lambda: self.window.controller.presets.editor.experts.change_selected()
        )
        self.setLayout(self.layout)

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def setup(self):
        """Setup layout"""
        layout = QHBoxLayout()
        arrows_layout = QVBoxLayout()
        self.window.ui.nodes["preset.experts.add"] = QPushButton(">")
        self.window.ui.nodes["preset.experts.add"].clicked.connect(
            lambda: self.window.controller.presets.editor.experts.add_expert()
        )
        self.window.ui.nodes["preset.experts.remove"] = QPushButton("<")
        self.window.ui.nodes["preset.experts.remove"].clicked.connect(
            lambda: self.window.controller.presets.editor.experts.remove_expert()
        )
        arrows_layout.addWidget(self.window.ui.nodes["preset.experts.add"])
        arrows_layout.addWidget(self.window.ui.nodes["preset.experts.remove"])

        self.window.ui.nodes["preset.experts.available.label"] = QLabel(trans("preset.experts.available.label"))
        self.window.ui.nodes["preset.experts.available"] = BaseList(self.window)
        self.window.ui.nodes["preset.experts.available"].clicked.disconnect()

        self.window.ui.nodes["preset.experts.selected.label"] = QLabel(trans("preset.experts.selected.label"))
        self.window.ui.nodes["preset.experts.selected"] = BaseList(self.window)
        self.window.ui.nodes["preset.experts.selected"].clicked.disconnect()

        available_layout = QVBoxLayout()
        available_layout.addWidget(self.window.ui.nodes["preset.experts.available.label"])
        available_layout.addWidget(self.window.ui.nodes["preset.experts.available"])

        selected_layout = QVBoxLayout()
        selected_layout.addWidget(self.window.ui.nodes["preset.experts.selected.label"])
        selected_layout.addWidget(self.window.ui.nodes["preset.experts.selected"])

        layout.addLayout(available_layout)
        layout.addLayout(arrows_layout)
        layout.addLayout(selected_layout)
        return layout

    def update_available(self, data):
        """
        Update available experts

        :param data: Data to update
        """
        id = "preset.experts.available"
        self.window.ui.nodes[id].backup_selection()
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[id].insertRow(i)
            name = data[n].name
            index = self.window.ui.models[id].index(i, 0)
            self.window.ui.models[id].setData(index, n, QtCore.Qt.ToolTipRole)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            i += 1
        self.window.ui.nodes[id].restore_selection()

    def update_selected(self, data):
        """
        Update selected experts

        :param data: Data to update
        """
        id = "preset.experts.selected"
        self.window.ui.nodes[id].backup_selection()
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[id].insertRow(i)
            name = data[n].name
            index = self.window.ui.models[id].index(i, 0)
            self.window.ui.models[id].setData(index, n, QtCore.Qt.ToolTipRole)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            i += 1
        self.window.ui.nodes[id].restore_selection()

    def update_lists(self, available: dict, selected: dict):
        """
        Update lists

        :param available: data with available experts
        :param selected: data with selected experts
        """
        self.update_available(available)
        self.update_selected(selected)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

import datetime
import os

from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans
from .experts import Experts


class Editor:
    def __init__(self, window=None):
        """
        Presets editor controller

        :param window: Window instance
        """
        self.window = window
        self.experts = Experts(window)
        self.options = {
            "filename": {
                "type": "text",
                "label": "preset.filename",
            },
            "name": {
                "type": "text",
                "label": "preset.name",
            },
            "ai_name": {
                "type": "text",
                "label": "preset.ai_name",
            },
            "user_name": {
                "type": "text",
                "label": "preset.user_name",
            },
            "img": {
                "type": "bool",
                "label": "preset.img",
            },
            "chat": {
                "type": "bool",
                "label": "preset.chat",
            },
            "completion": {
                "type": "bool",
                "label": "preset.completion",
            },
            "vision": {
                "type": "bool",
                "label": "preset.vision",
            },
            "langchain": {
                "type": "bool",
                "label": "preset.langchain",
            },
            "expert": {
                "type": "bool",
                "label": "preset.expert",
            },
            "agent": {
                "type": "bool",
                "label": "preset.agent",
            },
            # "assistant": {
            # "type": "bool",
            # "label": "preset.assistant",
            # },
            "llama_index": {
                "type": "bool",
                "label": "preset.llama_index",
            },
            "model": {
                "label": "toolbox.model.label",
                "type": "combo",
                "use": "models",
                "keys": [],
            },
            "temperature": {
                "type": "float",
                "slider": True,
                "label": "preset.temperature",
                "min": 0,
                "max": 2,
                "step": 1,
                "multiplier": 100,
            },
            "prompt": {
                "type": "textarea",
                "label": "preset.prompt",
                "context_options": ["prompt.template.paste"],
            },
            "tool.function": {
                "type": "dict",
                "label": "preset.tool.function",
                "keys": {
                    'name': 'text',
                    'params': 'textarea',
                    'desc': 'textarea',
                },
                "extra": {
                    "urls": {
                        "Help": "https://platform.openai.com/docs/guides/function-calling",
                    },
                },
            },
        }
        self.id = "preset"
        self.current = None

    def get_options(self) -> dict:
        """
        Get preset options

        :return: preset options
        """
        return self.options

    def get_option(self, id: str) -> dict:
        """
        Get preset option

        :param id: option id
        :return: preset option
        """
        return self.options[id]

    def setup(self):
        """
        Setup preset editor
        """
        # add hooks for config update in real-time
        self.window.ui.add_hook("update.preset.prompt", self.hook_update)

        # register functions dictionary
        parent = "preset"
        key = "tool.function"
        self.window.ui.dialogs.register_dictionary(
            key,
            parent,
            self.get_option(key),
        )

    def hook_update(self, key, value, caller, *args, **kwargs):
        """
        Hook: on settings update in real-time (prompt)
        """
        mode = self.window.core.config.get('mode')
        if key == "prompt":
            self.window.core.config.set('prompt', value)
            if mode == 'assistant':
                self.window.controller.assistant.from_global()  # update current assistant, never called!!!!!
            else:
                self.window.controller.presets.from_global()  # update current preset


    def edit(self, idx: int = None):
        """
        Open preset editor

        :param idx: preset index (row index)
        """
        preset = None
        if idx is not None:
            mode = self.window.core.config.get('mode')
            preset = self.window.core.presets.get_by_idx(idx, mode)
        self.init(preset)
        self.window.ui.dialogs.open_editor('editor.preset.presets', idx, width=800)

    def init(self, id: str = None):
        """
        Initialize preset editor

        :param id: preset id (filename)
        """
        data = PresetItem()
        data.name = ""
        data.filename = ""

        if id is not None and id != "":
            if id in self.window.core.presets.items:
                data = self.window.core.presets.items[id]
                data.filename = id
                self.current = data.uuid

        if data.name is None:
            data.name = ""
        if data.ai_name is None:
            data.ai_name = ""
        if data.user_name is None:
            data.user_name = ""
        if data.prompt is None:
            data.prompt = ""
        if data.filename is None:
            data.filename = ""

        # set current mode at start
        if id is None:
            mode = self.window.core.config.get("mode")
            if mode == "chat":
                data.chat = True
            elif mode == "completion":
                data.completion = True
            elif mode == "img":
                data.img = True
            elif mode == "vision":
                data.vision = True
            elif mode == "langchain":
                data.langchain = True
            # elif mode == "'"assistant':
                # data.assistant = True
            elif mode == "llama_index":
                data.llama_index = True
            elif mode == "expert":
                data.expert = True
            elif mode == "agent":
                data.agent = True

        options = {}
        data_dict = data.to_dict()
        for key in self.options:
            options[key] = self.options[key]
            options[key]['value'] = data_dict[key]

        # load options
        self.window.controller.config.load_options(
            self.id,
            options,
        )

        # update experts list, after ID loaded
        self.experts.update_list()

        # restore functions
        if data.has_functions():
            functions = data.get_functions()
            values = []
            for function in functions:
                values.append(
                    {
                        "name": function['name'],
                        "params": function['params'],
                        "desc": function['desc'],
                    }
                )
            self.window.ui.config[self.id]['tool.function'].items = values
            self.window.ui.config[self.id]['tool.function'].model.updateData(values)
        else:
            self.window.ui.config[self.id]['tool.function'].items = []
            self.window.ui.config[self.id]['tool.function'].model.updateData([])

        # set focus to name field
        self.window.ui.config[self.id]['name'].setFocus()

    def save(self, force: bool = False, close: bool = True):
        """
        Save ore create preset

        :param force: force overwrite file
        :param close: close dialog
        """
        id = self.window.controller.config.get_value(
            parent_id=self.id,
            key="filename",
            option=self.options["filename"],
        )
        mode = self.window.core.config.get("mode")
        modes = ["chat", "completion", "img", "vision", "langchain", "llama_index", "expert"]

        # disallow editing default preset
        if id == "current." + mode:
            self.window.ui.dialogs.alert("Reserved ID. Please use another ID.")
            return

        if id is None or id == "":
            name = self.window.controller.config.get_value(
                parent_id=self.id,
                key="name",
                option=self.options["name"],
            )
            if name is None or name == "":
                self.window.ui.dialogs.alert(trans('alert.preset.empty_id'))
                self.window.ui.status(trans('status.preset.empty_id'))
                return

            # generate new filename
            id = self.window.controller.presets.make_filename(name)
            path = os.path.join(
                self.window.core.config.path,
                "presets",
                id + ".json",
            )
            if os.path.exists(path) and not force:
                id = id + '_' + datetime.datetime.now().strftime('%Y%m%d%H%M%S')

        # validate filename
        id = self.window.controller.presets.validate_filename(id)
        if id not in self.window.core.presets.items:
            self.window.core.presets.items[id] = self.window.core.presets.build()
        elif not force:
            self.window.ui.dialogs.confirm(
                type='preset_exists',
                id=id,
                msg=trans('confirm.preset.overwrite'),
            )
            return

        # check if at least one mode is selected
        is_mode = False
        for check in modes:
            if self.window.controller.config.get_value(
                parent_id=self.id,
                key=check,
                option=self.options[check],
            ):
                is_mode = True
                break
        if mode != "agent" and not is_mode:
            self.window.ui.dialogs.alert(
                trans('alert.preset.no_chat_completion')
            )
            return

        # assign data from fields to preset object in items
        self.assign_data(id)

        # if agent, assign experts and select only agent mode
        if self.window.core.config.get('mode') == 'agent':
            self.window.core.presets.items[id].mode = ["agent"]

        # apply changes to current active preset
        current = self.window.core.config.get('preset')
        if current is not None and current == id:
            self.to_current(self.window.core.presets.items[id])
            self.window.core.config.save()

        # update current uuid
        self.current = self.window.core.presets.items[id].uuid

        # save
        self.window.core.presets.save(id)
        self.window.controller.presets.refresh()

        # close dialog
        if close:
            self.window.ui.dialogs.close('editor.preset.presets')
        else:
            # update ID in field
            self.window.controller.config.apply_value(
                parent_id=self.id,
                key="filename",
                option=self.options["filename"],
                value=id,
            )
        self.window.ui.status(trans('status.preset.saved'))

        # sort by name
        self.window.core.presets.sort_by_name()

        # switch to editing preset
        self.window.controller.presets.set(mode, id)

        # update presets list
        self.window.controller.presets.refresh()

    def assign_data(self, id: str):
        """
        Assign data from fields to preset

        :param id: preset id (filename)
        """
        data_dict = {}
        for key in self.options:
            # assigned separately
            if key == "tool.function":
                continue
            data_dict[key] = self.window.controller.config.get_value(
                parent_id=self.id,
                key=key,
                option=self.options[key],
            )
        if data_dict['name'] is None or data_dict['name'] == "":
            data_dict['name'] = id + " " + trans('preset.untitled')
        if data_dict['model'] == '_':
            data_dict['model'] = None

        preset = self.window.core.presets.items[id]
        preset.from_dict(data_dict)
        preset.tools = {
            'function': [],  # functions are assigned separately (below)
        }

        # assign functions tool
        values = self.window.controller.config.get_value(
            parent_id=self.id,
            key='tool.function',
            option=self.options['tool.function'],
        )
        functions = []
        for function in values:
            name = function['name']
            params = function['params']
            desc = function['desc']
            if name is None or name == "":
                continue
            if params is None or params == "":
                params = '{"type": "object", "properties": {}}'  # default empty JSON params
            if desc is None:
                desc = ""
            functions.append(
                {
                    "name": name,
                    "params": params,
                    "desc": desc,
                }
            )

        if len(functions) > 0:
            preset.tools['function'] = functions
        else:
            preset.tools['function'] = []

    def to_current(self, preset: PresetItem):
        """
        Update preset field from editor

        :param preset: preset item
        """
        self.window.core.config.set('ai_name', preset.ai_name)
        self.window.core.config.set('user_name', preset.user_name)
        self.window.core.config.set('prompt', preset.prompt)
        self.window.core.config.set('temperature', preset.temperature)

    def from_current(self):
        """Copy data from current active preset"""
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="ai_name",
            option=self.options["ai_name"],
            value=self.window.core.config.get('ai_name'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="user_name",
            option=self.options["user_name"],
            value=self.window.core.config.get('user_name'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="prompt",
            option=self.options["prompt"],
            value=self.window.core.config.get('prompt'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="temperature",
            option=self.options["temperature"],
            value=self.window.core.config.get('temperature'),
        )
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="model",
            option=self.options["model"],
            value=self.window.core.config.get('model'),
        )

    def update_from_global(self, key, value):
        """Update field from global config"""
        preset_id = self.window.core.config.get('preset')
        if preset_id is not None and preset_id != "":
            if preset_id in self.window.core.presets.items:
                preset = self.window.core.presets.items[preset_id]
                if key == 'preset.user_name':
                    preset.user_name = value
                    self.window.core.config.set('user_name', preset.user_name)
                elif key == 'preset.ai_name':
                    preset.ai_name = value
                    self.window.core.config.set('ai_name', preset.ai_name)
                self.window.core.presets.save(preset_id)
